import copy
import datetime
import os
import re

import numpy as np
import pandas as pd

from yangke.base import save_as_xlsx, cut_time_period_of_dataframe, get_datetime_col_of_dataframe, is_number, \
    is_contain_chinese, is_datetime, set_dataframe_header, merge_two_dataframes
from yangke.common import fileOperate
from yangke.common.config import logger
from yangke.common.fileOperate import read_csv_ex, write_lines
from yangke.performance.dp.file_type_detect import get_detailed_type_of_data_file
from yangke.performance.dp.const_name import suffix_p_10, suffix_dcs_10, suffix_imp_20, suffix_dcs_20, suffix_p_20, \
    suffix_imp_10, suffix_p_01, suffix_dcs_01


def get_description_row(df):
    """
    获取pandas的dataframe中表示测点名称的行，一般测点名称为Description

    :param df:
    :return:
    """
    header = min(len(df), 4)
    for i in range(header):
        for j in range(len(df.iloc[i])):
            if (is_contain_chinese(str(df.iloc[i, j])) and "时间" not in str(df.iloc[i, j])) \
                    or " " in str(df.iloc[i, j]):
                return i


def get_kks_row(df):
    """
    获取pandas的dataframe中表示kks码的行

    :param df: dataframe数据
    :return:
    """
    possible_row = [0, 1, 2, 3]
    header = min(len(df), len(possible_row))
    for i in range(header):
        for j in range(len(df.iloc[i])):
            value = str(df.iloc[i, j])
            if value == 'nan' or value == '':
                continue
            if is_number(value):  # kks码不可能是纯数字，因此如果是纯数值，则说明该行不是kks码
                possible_row.remove(i)
                break  # 进行下一行
            elif is_contain_chinese(value) and not ("时间" in str(df.iloc[i, j])):  # kks码不可能包含中文汉字，因此包含中文，则说明该行不是kks码
                possible_row.remove(i)
                break  # 进行下一行

    if len(possible_row) == 1:
        return possible_row[0]
    elif len(possible_row) == 0:
        return None
    logger.warning("未能确定KKS码所在的行")
    return None


def power_file_pre_deal(file):
    """
    功率原始文件预处理：
    1.删除功率文件的表头行，只保留数据行
    2.合并功率表文件中date和time列

    :param file:
    :return:
    """
    try:
        data = read_csv_ex(file, sep="\t", header=None)
        start_line = 0
        for i, line in enumerate(data.values):
            line = line[0].strip()
            if line.startswith("Store No"):
                start_line = i
                break
        lines = [line[0].replace('"', '') for line in data.iloc[start_line:, :].values]
        outfile1 = os.path.join(os.path.dirname(file),
                                os.path.splitext(os.path.basename(file))[0] + f"{suffix_p_01}.csv")
        write_lines(outfile1, lines)
        data = read_csv_ex(outfile1)

        if data.columns[1] == "Date" and data.columns[2] == "Time":
            data["Date"] = data["Date"] + " " + data["Time"]
            data.rename(columns={"Date": "DateTime"}, inplace=True)
            data = data.drop(columns="Time")
            data.to_csv(outfile1, index=None)
        else:
            logger.warning(f"功率文件{file}的第2列和第三列数据不是‘Date’和‘Time’，请检查文件！")
            return None
        return outfile1
    except PermissionError:
        logger.error("无法写入文件，请检查目标文件是否占用！")


def sis_file_pre_deal(file):
    """
    预处理SIS系统导出的文件，将测点名称放置到第一行，删除首行以外的其他非数据行

    该方法会将任何数据文件中首行以外的非数据行删除，例如单位行、kks码行、表号行等删除
    该方法会保留测点名诚行，如果测点名称行不在第一行，该方法会识别测点名称行并将测点名称行移到第一行

    :param file:
    :return:
    """
    data = read_csv_ex(file, sep=",", header=None, index_col=0)
    row_des = get_description_row(data)
    row_kks = get_kks_row(data)
    data.loc["时间"] = data.iloc[row_des]
    data.rename(index={"时间": "DateTime"}, inplace=True)
    # --------------------------- 删除非数据和非测点名称行 -------------------------------------------
    drop_lines = []
    for i in range(1, min(len(data), 10)):
        col1_val = str(data.iloc[i, 0])
        if not is_datetime(col1_val) and not is_number(col1_val):
            drop_lines.append(i)
    data = data.reset_index().drop(drop_lines).set_index(0)  # 删除kks码所在的行
    # --------------------------- 删除非数据和非测点名称行 -------------------------------------------
    out_file = os.path.join(os.path.dirname(file), os.path.basename(file)[:-4] + suffix_dcs_01 + ".csv")
    data.to_csv(out_file, header=None)
    return out_file


def dcs_hollysys_file_pre_deal(file):
    """
    和利时DCS文件预处理，只需要更改时间列的列名（【时间列】->【DateTime】）
    :param file:
    :return:
    """

    def strip(text):
        try:
            return text.strip()
        except AttributeError:
            return text

    def add_prefix(string):
        return f"20{string}"

    outfile = os.path.join(os.path.dirname(file),
                           os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")
    # ---------- 删除文件中的\t和前后的空格，不删除会导致将数值读取为文本 --------------------
    data = read_csv_ex(file, low_memory=False, header=None)
    data = data.applymap(strip)  # 对data的每一个元素应用strip方法
    data.dropna(axis=1, how="all", inplace=True)  # 先去除空白字符，再丢弃空行，否则空白有可能被认为是有内容的
    data.dropna(axis=0, how="all", inplace=True)
    data.to_csv(outfile, index=False, header=False)
    # ---------- 删除文件中的\t，不删除会导致将数值读取为文本 --------------------
    data = read_csv_ex(outfile, low_memory=False)
    data.rename(columns={"时间段": "DateTime"}, inplace=True)
    data["DateTime"] = data["DateTime"].map(add_prefix)
    try:
        data.to_csv(outfile, index=False)
        return outfile
    except PermissionError:
        logger.error("无法写入文件，请检查目标文件是否占用！")


last_time = None  # 不能删除


def dcs_tpri_file_pre_deal(file):
    """
    处理热工院DCS导出的数据文件，该文件有多个时间列，需要特殊处理

    :param file:
    :return:
    """
    outfile = os.path.join(os.path.dirname(file),
                           os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")

    global last_time
    temp_ = None

    def change_title(text):
        """
        将DCS文件中的两个标题行合并成一行，使其变成以日期和点描述为标题的数据DataFrame

        :param text:
        :return:
        """
        global temp_  # 不能删除，否则内部函数会认为该变量是内部定义的变量，导致不适用外部的该参数
        if text == np.nan:
            return None
        if type(text) == str:
            text: str = text.strip()
            if "时间" in text:
                temp_ = text[:-2]
                return "DateTime"
            elif text.endswith("值"):
                if temp_ is None:
                    return text[:-1]
                else:
                    return text[:-1] + f"_{temp_}"
            elif "状态" in text:
                return "状态"
        else:
            return "状态"

    def add_date_to_time(text):
        """
        向原始数据的时间列添加日期，需要处理跨天的情况

        :param text:
        :return:
        """
        global last_time
        last_time_date_str = datetime.datetime.strftime(last_time, "%Y-%m-%d")
        current_time_str = f"{last_time_date_str} {text}"
        current_time = datetime.datetime.strptime(current_time_str, "%Y-%m-%d %H:%M:%S")
        if current_time < last_time:  # 说明跨天了，更新日期
            current_time = current_time + datetime.timedelta(days=1)
            last_time = current_time
            current_time_str = datetime.datetime.strftime(current_time, "%Y-%m-%d %H:%M:%S")
        return current_time_str

    data = read_csv_ex(file)
    # ------------------------- 获取数据记录日期 -----------------------------------
    cell = data.columns[0]
    date_list = re.findall("历史趋势开始于(\d{4})年(\d{1,2})月(\d{1,2})日(\d{1,2})时(\d{1,2})分(\d{1,2})秒", cell)
    if len(date_list) > 0:
        year = int(date_list[0][0])
        month = int(date_list[0][1])
        day = int(date_list[0][2])
        hour = int(date_list[0][3])
        minute = int(date_list[0][4])
        second = int(date_list[0][5])
        start_time = datetime.datetime(year, month, day, hour, minute, second)
        last_time = copy.deepcopy(start_time)  # 记录上一条数据的时间
    else:
        logger.debug("找不到工况记录的日期")
        return None
    # ------------------------- 获取数据记录日期 -----------------------------------
    row1: pd.Series = data.iloc[0]
    row1.fillna("", inplace=True)
    row2: pd.Series = data.iloc[1]
    row2.fillna("", inplace=True)
    title_series: pd.Series = row1 + row2
    title_series = title_series.map(change_title)
    title_series[0] = "序号"
    if data.shape[1] <= 2:
        logger.info(f"文件{file}中的数据量为0")
        return None
    data.drop(index=[0, 1], inplace=True)
    data.set_axis(axis=1, labels=title_series, inplace=True)
    # 删除状态列，删除重复列，删除某些人处理的平均行
    data: pd.DataFrame = data.T.drop(index=["状态", "序号"]).dropna(axis=1, how="any").drop_duplicates().T
    data["DateTime"] = data["DateTime"].map(add_date_to_time)
    data.to_csv(outfile, index=False)
    return outfile


def dcs2_file_pre_deal(data: pd.DataFrame, file):
    """
    处理dcs2格式的dataframe

    :param data:
    :return:
    """
    outfile = os.path.join(os.path.dirname(file),
                           os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")
    # ------------------------- 找到DCS文件中表头描述所在的区域 ------------------------------
    start1 = 0  # kks码和点位名称区域的开始行
    end1 = 0  # kks码和点位名称区域的结束行
    start2 = 0  # 数据区域的开始行
    during_point_description_section = False
    for i, line in enumerate(data.values):
        cell00 = str(line[0]).replace('"', '').strip()  # 将ndarray转换为字符串

        if during_point_description_section:  # 需要在空行判断之前进行
            try:
                _ = int(cell00)
                continue
            except ValueError:
                end1 = i
                during_point_description_section = False
                continue
        if cell00 == "" or cell00 == "nan":  # 元素为空
            continue

        if cell00.strip() == "1" and end1 == 0:
            start1 = i - 1
            during_point_description_section = True
        else:
            cell0 = cell00.strip().split(" ")[0]
            try:  # 找到第一个首元素可以转为日期的行
                _ = pd.Timestamp(str(cell0))
                start2 = i
                break
            except ValueError:
                continue
    if end1 == 0 or start2 == 0:
        return None
    # ---------------------- 找到DCS文件中表头描述所在的区域 ------------------------------
    # ---------------------- 将data分割为表头区域的DataFrame和数据区域的DataFrame -------------------------
    title_df = data[start1:end1]
    data_df = data[start2:].copy()
    data_df.dropna(axis=1, how="all", inplace=True)
    # ---------------------- 将data分割为表头区域的DataFrame和数据区域的DataFrame -------------------------
    title_list = list(set_dataframe_header(title_df)["Description"])
    title_list.insert(0, "DateTime")
    try:
        data_df = data_df.set_axis(title_list, axis=1, copy=False)  # 新版本语法
    except:
        data_df.set_axis(title_list, axis=1, inplace=True)  # 旧版本语法

    data_df.to_csv(outfile, sep=",", index=None)
    return outfile


def dcs_file_pre_deal(file):
    """
    dcs原始文件预处理：
    返回预处理得到的csv类型的文件的文件名

    :param file:
    :return:
    """
    try:
        outfile = os.path.join(os.path.dirname(file),
                               os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")
        ext_ = os.path.splitext(file)[-1].lower()
        sep = ","
        if ext_ == ".txt":
            sep = " "
        elif ext_ == ".csv":
            sep = ","
        data = read_csv_ex(file, sep=sep, header=None)  # sep不能取值"/t"，否则读取dcs1.csv格式时报错，可能与to_csv方法行为改变有关
        data.dropna(axis=1, how="all", inplace=True)
        data.to_csv(outfile, sep=",", index=False, header=False)  # 写出当前data到新文件，可以删除源文件中的空行
        start1 = 0  # kks码和点位名称区域的开始行
        end1 = 0  # kks码和点位名称区域的结束行
        start2 = 0  # 数据区域的开始行
        for i, line in enumerate(data.values):
            line1 = str(line[0]).replace('"', '')
            line2 = str(line[1]).replace('"', '')
            if line1 == "nan" and line2.startswith("Graph Visible"):
                start1 = i
            elif line1.strip().startswith("Date Time"):
                end1 = i - 1
            else:
                cell0 = line[0]
                if line1 == "nan":
                    continue
                try:  # 找到第一个首元素可以转为日期的行
                    _ = pd.Timestamp(str(cell0))
                    start2 = i
                    break
                except ValueError:
                    continue

        data_title_field = read_csv_ex(outfile, sep=",", skiprows=range(start1), nrows=end1 - start1)
        point_name = list(data_title_field["Description"])  # dcs测点名
        point_name.insert(0, "DateTime")
        data_value_field = read_csv_ex(outfile, sep=",", skiprows=range(start2 - 1))
        data_value_field.columns = point_name
        data_value_field.sort_values(by="DateTime", inplace=True)  # 写出数据前对时间进行排序
        data_value_field.to_csv(outfile, sep=",", index=False)
        return outfile
    except PermissionError:
        logger.error("无法写入文件，请检查目标文件是否占用！")


def dcs_sciyon_file_pre_deal(file):
    """
    预处理南京科远DCS系统导出的数据文件： *.csv
    :param file:
    :return:
    """
    outfile = os.path.join(os.path.dirname(file),
                           os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")
    data: pd.DataFrame = read_csv_ex(file)  # 跳过前三行，前三行是文件信息描述，不是数据区域
    start_row = 2
    for i in range(data.shape[0]):
        if data.iloc[i][0].startswith('$TagDesc'):
            start_row = i + 1  # 第一行已经被识别为标题，因此开始行=i+1
            break
    data = read_csv_ex(file, skiprows=start_row)
    data.drop(index=[0], inplace=True)  # 删除第一行，该行描述了标签的格式，也不是数据区域
    data.rename(columns={"$TagDesc": "DateTime"}, inplace=True)  # 重命名时间列
    data['DateTime'] = data['DateTime'].apply(lambda x: (x.replace("@", "")))  # 删除时间之前的@符号
    data = data.applymap(lambda x: (x.replace("(GOOD)", "")))  # 删除数据后的质量标签
    data = data.applymap(lambda x: (x.replace("(BAD)", "")))
    try:
        data.to_csv(outfile, sep=",", index=False)
    except PermissionError:
        logger.error("无法写入文件，请检查目标文件是否占用！")
    return outfile


def sppa_t3000_pre_deal(file, sep):
    def deal_page(p: pd.DataFrame):
        types_p = ""
        p = p.reset_index().copy()
        del p[p.columns[0]]  # 删除第一列
        for i, r in p.iterrows():
            r = r.fillna(" ").astype(str)
            line_ = " ".join(r.tolist()).strip()
            if line_.__contains__("Name") and line_.__contains__("Designation"):
                # 说明是表头说明页
                types_p = "说明页"
                break
            elif r[1] == "Time" or r[0] == "Time":
                # 说明是数据页面
                types_p = "数据页"
                break
            else:
                continue

        if types_p == "说明页":
            _s = 0
            _e = 0
            for i, r in p.iterrows():
                line_ = " ".join(r.tolist()).strip()
                if line_.__contains__("Designation"):
                    _s = i
                elif line_.__contains__("Page") and line_.__contains__("of"):
                    _e = i
            p: pd.DataFrame = p.iloc[_s:_e].copy()
            p = p.reset_index(drop=True)  # 给行重新编号
            p = p.set_axis(p.iloc[0], axis=1)  # 以第0行为列标题
            p = p.drop(index=[0])
            p = p.replace(" ", np.nan).dropna(axis=1, how="all")
            if p.columns[0] == ' ':
                p = p.rename({" ": "tagId"}, axis=1)
            tags = {}
            for (i, name, des, unit) in zip(p["tagId"], p["Name"], p["Designation"], p["EngUnit"]):
                tags.update({f"Tag{i}": f"Tag{i}_{des}_{name}_{unit}"})
            return tags
        elif types_p == "数据页":
            _s = 0
            _e = 0
            for i, r in p.iterrows():
                line_ = " ".join(list(r)).strip()
                if "Time" in line_ and "Tag" in line_:
                    _s = i
                elif line_.__contains__("Page") and line_.__contains__("of"):
                    _e = i
            p = p.iloc[_s: _e].copy()
            p = p.replace(to_replace=" ", value=np.nan).dropna(how="all").dropna(how="all", axis=1)
            p = p.set_axis(p.iloc[0], axis=1)
            p = p.reset_index(drop=True).drop([0, 1], axis=0)
            p = p.reset_index(drop=True)
            return p
            # 将单位与标题列合并

    outfile = os.path.join(os.path.dirname(file),
                           os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")
    df = read_csv_ex(file, sep=sep)
    df = df.fillna(" ").astype(str)
    pages = []
    _start = 0
    _end = 0
    for idx, row in df.iterrows():
        line = " ".join(list(row)).strip()
        if line.startswith('Analog Grid Report'):
            _end = idx
            if _end != _start:
                page = df.iloc[_start: _end].copy()
                pages.append(page)
                _end = 0  # 一次分割完成后，需要将_end重置为0，否则_end就等于_start了
            _start = idx
    pages.append(df.iloc[_start:])  # 剩下的部分是最后一页
    data_total: pd.DataFrame | None = None
    data_total_same_tag: pd.DataFrame | None = None
    tags_all = {}
    for page in pages:
        res = deal_page(page)
        if isinstance(res, dict):
            tags_all.update(res)
        else:
            if data_total_same_tag is None:
                data_total_same_tag = res
            else:
                if set(data_total_same_tag.columns) == set(res.columns):  # 说明是同一批参数不同时间的数据页
                    data_total_same_tag, _ = merge_two_dataframes(data_total_same_tag, res)
                else:  # 说明是不同批参数的数据页，此时，上一批参数的数据页结束，data_total_same_tag中储存了上一批参数所有时间段的数据
                    # 将上一批参数所有时间段的数据保存到data_total中，然后将data_total_same_tag重置为None，开始进行下一批参数的处理
                    if data_total is None:
                        data_total = data_total_same_tag.copy()
                    else:
                        data_total, _ = merge_two_dataframes(data_total, data_total_same_tag)

                    data_total_same_tag = res.copy()

    if data_total is None:  # 如果只有一批数据，则data_total为None，data_total_same_tag收集的就是所有数据
        data_total = data_total_same_tag
    else:  # 如果有多批数据，需要将最后一批data_total_same_tag收集的数据添加到data_total中
        data_total, _ = merge_two_dataframes(data_total, data_total_same_tag)

    logger.debug(f"输出预处理修正文件")
    data_total = data_total.rename(columns={"Time": "DateTime"}).rename(columns=tags_all)  # 重命名列名

    try:
        data_total.to_csv(outfile, sep=",", index=False)
    except PermissionError:
        logger.error("无法写入文件，请检查目标文件是否占用！")
    return outfile


def deal_dcs_pre_file(file, time_dict):
    """
    处理dcs预处理得到的csv文件，该文件是由该软件生成的中间文件，所有不同的dcs系统都将转为统一的格式，格式示例如下：
    DateTime,左侧主汽压力1,左侧主汽压力2
    2021-12-17 09:00:00,12.11,11.99
    2021-12-17 09:00:12,12.11,11.29

    :param file:
    :param time_dict: 工况确认单里的各工况时间，是一个字典，格式为
    {condition_name: {"start_time": datetime_str, "end_time": datetime_str}}
    :return:
    """
    try:
        if file is None:
            logger.warning("dcs数据处理失败！")
            return None
        else:
            data = read_csv_ex(file, index_col=0)
            conditions = []
            out_files = []
            for condition_name, start_end_time_dict in time_dict.items():
                start_time = start_end_time_dict.get("start_time")
                end_time = start_end_time_dict.get("end_time")
                data1 = cut_time_period_of_dataframe(data, start_time, end_time)
                if len(data1) > 0:
                    out_file2 = os.path.join(os.path.dirname(file), f"temp_{condition_name}{suffix_dcs_10}.csv")
                    out_files.append(out_file2)
                    data1 = data1.sort_values(by="DateTime")  # 根据DateTime对数据排序
                    data1.to_csv(out_file2)
                    conditions.append(condition_name)
            return "dcs_slice", out_files, conditions
    except PermissionError:
        logger.error("文件写入失败，请检查目标文件是否被占用！")


def unified_unit(df: pd.DataFrame, unit_series):
    """
    统一df中各个测点的单位，统一后的单位为 压力-kPa，温度-℃，流量差压-kPa

    :param df:
    :param unit_series:
    :return:
    """
    for col, unit in enumerate(unit_series):
        if str(unit).lower().strip() == "mpa":
            df.iloc[:, col] = df.iloc[:, col] * 1000
    return df


def is_unit_series(series: pd.Series):
    """
    判断给定的Series是不是表示单位的Series

    :param series:
    :return:
    """
    for index, value in series.items():
        value = str(value).strip()
        if value == "" or value == "nan":
            continue
        if value.lower() in ["kpa", "℃", "mm", "m", "mpa", "t/h", "kg/s", "v", "a", "kw"]:
            return True
    return False


def cutting_period_of_dataframe(df, time_dict):
    """
    依次查询time_dict中的时间段是否在df数据集中，如果存在返回对应时间段的子数据集，则记录该数据集在time_dict中的key和其对应的子数据集，如果
    所有的time_dict都不存在，则返回None

    :param df:
    :param time_dict:
    :return: {"工况名": pd.DataFrame, ...}，如果未找到任何工况数据，返回None
    """

    name_data_dict = {}
    for condition_name, start_end_time_dict in time_dict.items():
        start_time = start_end_time_dict.get("start_time")
        end_time = start_end_time_dict.get("end_time")
        data1 = cut_time_period_of_dataframe(df, start_time, end_time)
        if len(data1) > 0:
            name_data_dict.update({condition_name: data1})
    return name_data_dict


def deal_kp_hollysys(data):
    point_description = {}
    for sheet_name, content in data.items():
        # 判断content是否包含PN和DS两列，分别代表KKS码和点描述
        content: pd.DataFrame = set_dataframe_header(content)  # 将第一行转为Header，这一行是英文标题
        content.reset_index(inplace=True)  # 将Index转为列
        if "PN" in content.columns and "DS" in content.columns:
            pass
        else:
            continue  # 略过后续操作处理下一个表单
        content.drop(content.index[0], inplace=True)  # 删除第二行，第二行是中文标题

        for pn, ds in zip(content["PN"], content["DS"]):
            point_description[pn] = ds
    return point_description


def pre_deal_and_cutting_period(file, time_dict, bound, config: dict):
    """
    预处理文件，如果文件中包含“时间-数据”条目，则剪切到对应工况的时间段，返回文件类型及剪切后的文件
    从file中裁取指定时间段的试验数据，并剔除数据列中的错误数据

    :param time_dict:
    :param file:
    :return: 返回第一个参数为数据文件类型，可能取值"imp_slice"/"power_slice"/"dcs_slice"，第二个参数为生成的预处理文件列表，第三个参数为
    预处理的工况名称列表，如果出错，则返回None
    """
    if os.path.splitext(os.path.basename(file))[0].endswith("修正"):  # 只处理原始文件，不处理修正数据
        return None
    ext_ = os.path.splitext(file)[-1].lower()
    type_ = get_detailed_type_of_data_file(file)
    if ext_ == ".xls":  # 老版本的excel文件，另存为新版本的再处理
        file = save_as_xlsx(file, ext="csv", engine="WPS")
        pre_deal_and_cutting_period(file, time_dict, bound)
    if ext_ == ".xlsx":
        # =========================== 读入现有的采集数据，构造本步骤的输出文件名 ======================================
        data = pd.read_excel(file, header=None, sheet_name=None)
        conditions = []
        out_file_list = []
        out_file_dir = os.path.join(os.path.dirname(file), "imp")
        os.makedirs(out_file_dir, exist_ok=True)
        # =========================== 读入现有的采集数据，构造本步骤的输出文件名 ======================================
        if type_ == "imp":
            # =========================== 有线采集仪数据处理 ======================================
            for sheet_name, df in data.items():
                if df.shape != (0, 0):
                    name_series = df.iloc[1, :].str.strip()
                    tag_series = df.iloc[2, :].str.strip()
                    tag_series.replace(to_replace=np.nan, value="", inplace=True)
                    unit_series = df.iloc[3, :].str.strip()
                    unit_series.replace(to_replace=np.nan, value="", inplace=True)
                    name_series[1:] = name_series[1:] + "_" + unit_series[1:] + "_" + tag_series[1:]
                    data1 = df.set_axis(labels=name_series, axis='columns')
                    data_no_title = data1.drop(index=[0, 1, 2, 3])
                    data_no_title.rename(columns={"名称": "DateTime"}, inplace=True)
                    logger.info("sheet: " + sheet_name + "处理中...")
                    data_no_title = unified_unit(data_no_title, unit_series)
                    cut_result = cutting_period_of_dataframe(data_no_title, time_dict)
                    if cut_result is not None:
                        imp_tested = config.get("imp_or_wsn_tested")
                        for condition_name, current_data in cut_result.items():
                            out_file = os.path.join(out_file_dir, f"{condition_name}{suffix_imp_10}.csv")
                            current_data = current_data.sort_values(by="DateTime")  # 写出数据前对时间排序
                            # 是否有其他原始数据文件中存在当前工况数据
                            exist_in_whole = imp_tested is not None and condition_name in imp_tested
                            if exist_in_whole:  # 有则合并当前工况数据和其他原始数据文件中的数据
                                _exist = read_csv_ex(out_file)
                                current_data = merge_two_dataframes(current_data, _exist)[0]
                            current_data.to_csv(out_file, index=None)
                            if out_file not in out_file_list and not exist_in_whole:  # 如果每输出过，则输出
                                out_file_list.append(out_file)
                            if condition_name not in conditions and not exist_in_whole:
                                conditions.append(condition_name)
                            else:
                                logger.warning("imp采集数据单个工况对应的时间段存在多个文件中")
            return "imp_slice", out_file_list, conditions
            # =========================== 有线采集仪数据处理 ======================================
        elif type_ == "wsn" or type_ == "dcs_pi":
            # =========================== 无线采集仪数据处理 ======================================
            for sheet_name, df in data.items():
                if df.shape != (0, 0):
                    name_series = df.iloc[0].str.strip()
                    unit_series: pd.Series = df.iloc[1].str.strip()
                    if type_ == "dcs_pi":
                        if not is_unit_series(unit_series):
                            unit_series = df.iloc[2].apply(str).str.strip()
                        if not is_unit_series(unit_series):
                            logger.warning("dcs_pi数据未找到单位行，默认使用第三行作为单位行")
                    else:
                        # 这里使用series.str必须先经过series.apply(str)转换，否则，当series全是数值类型时会报错如下：
                        # {AttributeError}Can only use .str accessor with string values!
                        tag_series: pd.Series = df.iloc[2].apply(str).str.strip()
                        tag_series.replace(to_replace=np.nan, value="", inplace=True)
                        unit_series[1:] = unit_series[1:] + "_" + tag_series[1:]
                    unit_series.replace(to_replace=np.nan, value="", inplace=True)
                    name_series[1:] = name_series[1:] + "_" + unit_series[1:]
                    data_no_title: pd.DataFrame = df.drop(index=[0, 1, 2])
                    data_no_title = data_no_title.set_axis(labels=name_series, axis='columns')
                    data_no_title.replace(to_replace="INVALID", value=np.nan, inplace=True)
                    if type_ == "wsn":  # wsn需要单独处理时间列
                        year_str = ""
                        for condition, start_end_time in time_dict.items():
                            start_time_str = start_end_time.get("start_time")
                            if year_str != "":
                                if start_time_str[0:4] != year_str:
                                    logger.warning("工况确认单中数据跨年了，暂时不支持跨年数据处理")
                            year_str = start_time_str[0:4]
                        if year_str == '':
                            logger.error("wsn年份查询失败，可能是工况记录单未找到或工况记录单中条目数为0或数据目录错误")
                        data_no_title["时间"] = year_str + "." + data_no_title["时间"]
                        data_no_title["时间"] = pd.to_datetime(data_no_title['时间'], format="%Y.%m.%d %H:%M:%S")
                        data_no_title.rename(columns={"时间": "DateTime"}, inplace=True)
                        data_no_title = unified_unit(data_no_title, unit_series)  # 统一数据集中测点数据的单位
                    else:  # pi数据库导出的dcs数据
                        time_col = get_datetime_col_of_dataframe(data_no_title)
                        data_no_title.rename(columns={time_col: "DateTime"}, inplace=True)
                    name_data_dict = cutting_period_of_dataframe(data_no_title, time_dict)  # 裁剪对应工况时间段
                    if len(name_data_dict) > 0:
                        conditions.extend(name_data_dict)
                        for condition, c_df in name_data_dict.items():
                            condition: str = condition.strip()
                            if type_ == "wsn":
                                temp_suffix = suffix_imp_10
                            else:
                                temp_suffix = suffix_dcs_10
                            c_out_file = os.path.join(out_file_dir,
                                                      "temp_" + condition + f"{temp_suffix}.csv")
                            c_df = c_df.sort_values(by="DateTime")  # 写出数据前对时间排序
                            c_df.to_csv(c_out_file, index=None)
                            out_file_list.append(c_out_file)
            if type_ == "wsn":
                return "wsn_slice", out_file_list, conditions
            else:
                return "dcs_slice", out_file_list, conditions
            # =========================== 无线采集仪数据处理 ======================================
        elif type_ == "dcs2":  # .xls格式的DCS数据文件会在预处理时转换为.xlsx格式，然后会返回dcs2类型
            for sheet_name, df in data.items():
                if df.shape != (0, 0):
                    df.dropna(axis=0, how="all")
                    df.dropna(axis=1, how="all")
                    outfile1 = dcs2_file_pre_deal(df, file)  # 将原始dcs文件转换为统一的中间格式
                    if outfile1 is None:
                        logger.info(f"DCS文件不包含任何可识别数据，文件路径为：{file}")
                        return None
                    return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "kp_hollysys":
            # 和利时的点表文件
            point_description = deal_kp_hollysys(data)
            return "kp_hollysys", point_description
    elif ext_ == ".csv":  # 功率数据一般为CSV
        if type_ == "power":
            outfile1 = power_file_pre_deal(file)
            if outfile1 is None:
                logger.warning("功率数据处理失败！")
                return None
            else:
                data = read_csv_ex(outfile1, index_col=0)
                name_data_dict = cutting_period_of_dataframe(data, time_dict)
                if name_data_dict is None:
                    return None
                out_files = []
                folder = os.path.join(os.path.dirname(file), "power")
                os.makedirs(folder, exist_ok=True)
                conditions = []
                for name, data in name_data_dict.items():
                    outfile1 = os.path.join(folder, f"{name}{suffix_p_10}.csv")
                    data = data.sort_values(by="DateTime")  # 写出数据前对时间排序
                    data.to_csv(outfile1)
                    out_files.append(outfile1)
                    conditions.append(name)
                return "power_slice", out_files, conditions
        elif type_ == "dcs1":
            outfile1 = dcs_file_pre_deal(file)  # 将原始dcs文件转换为统一的中间格式
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "dcs2":
            df = read_csv_ex(file)
            outfile1 = dcs2_file_pre_deal(df, file)  # 将原始dcs文件转换为统一的中间格式
            if outfile1 is None:
                logger.info(f"DCS文件不包含任何可识别数据，文件路径为：{file}")
                return None
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "dcs_hollysys":
            # ----------------------------- 和利时的DCS数据测点标题是KKS码，需要电表进一步处理为点描述 --------------------------
            outfile1 = dcs_hollysys_file_pre_deal(file)  # 将原始dcs文件转换为统一的中间格式
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "dcs_tpri":
            outfile1 = dcs_tpri_file_pre_deal(file)
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "dcs_sciyon":
            outfile1 = dcs_sciyon_file_pre_deal(file)
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "sis":
            outfile1 = sis_file_pre_deal(file)
            if outfile1 is None:
                logger.warning("DCS数据处理失败！（SIS）")
                return None
            else:
                data = read_csv_ex(outfile1, index_col=0)
                name_data_dict = cutting_period_of_dataframe(data, time_dict)
                if name_data_dict is None:
                    return None
                out_files = []
                folder = os.path.join(os.path.dirname(file), "dcs")
                os.makedirs(folder, exist_ok=True)
                conditions = []
                for name, data in name_data_dict.items():
                    outfile1 = os.path.join(folder, f"{name}{suffix_dcs_01}.csv")
                    data = data.sort_values(by="DateTime")
                    data.to_csv(outfile1)
                    out_files.append(outfile1)
                    conditions.append(name)
                return "dcs_slice", out_files, conditions
        elif type_ == "SPPA-T3000 and ; sep":
            outfile1 = sppa_t3000_pre_deal(file, ";")
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "SPPA-T3000 and , sep":
            outfile1 = sppa_t3000_pre_deal(file, ",")
            logger.info(f"暂不支持{type_=}的数据文件")
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "unknown":
            # 未知格式的文件，尝试按时间分割文件
            logger.info(f"未识别的文件类型，{file}")
            data = read_csv_ex(file)
            time_col_name = get_datetime_col_of_dataframe(data)
            if time_col_name is None:
                return
            data.rename(columns={time_col_name: "DateTime"}, inplace=True)
            out_file = os.path.join(os.path.dirname(file),
                                    os.path.basename(file)[:-4] + "_unk预处理_修正.csv")
            title_rows = []
            for i in range(10):  # 判断前10行是数据行还是标题行，如果是标题信息行，则合并各行
                cell0 = data["DateTime"].iloc[i]
                try:  # 找到第一个首元素可以转为日期的行
                    if str(cell0).strip() == "":
                        continue
                    _ = pd.Timestamp(str(cell0))
                    break
                except ValueError:
                    title_rows.append(i)
            if len(title_rows) > 0:
                title = data.columns
                for i in title_rows:
                    title = title + "_" + data.iloc[i]
                data.set_axis(title, axis=1, inplace=True)

            time_col_name = get_datetime_col_of_dataframe(data)
            data.rename(columns={time_col_name: "DateTime"}, inplace=True)
            data.to_csv(out_file, index=None)
            name_data_dict = cutting_period_of_dataframe(data, time_dict)
            if name_data_dict is None:
                return None
            out_files = []
            folder = os.path.join(os.path.dirname(out_file), "UNK")
            os.makedirs(folder, exist_ok=True)
            conditions = []
            discriminate_unk_file = config.get("未知类型文件区分文件名")  # 在点名中是否添加文件名进行区分

            for name, data in name_data_dict.items():
                outfile1 = os.path.join(folder, f"{name}_unk预处理_修正.csv")
                data.dropna(axis=1, how="all", inplace=True)
                data = data.sort_values(by="DateTime")
                if discriminate_unk_file:
                    # 将文件名添加到列明上，因为位置类型文件的文件名可能是数据点的名称
                    name_map = {item: f"{os.path.basename(file)[:-4]}-{item}" for item in list(data.columns) if
                                item != "DateTime"}
                    data.rename(columns=name_map, inplace=True)
                data.to_csv(outfile1, index=None)
                out_files.append(outfile1)
                conditions.append(name)
            return "unk_slice", out_files, conditions
    elif ext_ == ".txt":
        if type_ == "dcs1":
            outfile1 = dcs_file_pre_deal(file)  # 将原始dcs文件转换为统一的中间格式
            return deal_dcs_pre_file(outfile1, time_dict)
        elif type_ == "dcs2":
            # 难以使用dcs2_file_pre_deal，这里单独处理
            outfile1 = os.path.join(os.path.dirname(file),
                                    os.path.splitext(os.path.basename(file))[0] + suffix_dcs_01 + ".csv")
            lines = fileOperate.read_lines(file, dropna=True)  # 删除文件中的空行，以防pandas出现莫名其妙的问题
            write_lines(outfile1, lines)
            start1 = 0
            end1 = 0
            start2 = 0
            fix_end1 = False  # 是否允许end1变化
            for i in range(len(lines)):
                line = lines[i].strip()
                if not line:  # 跳过空行
                    continue
                if line.startswith("G"):
                    start1 = i
                    end1 = 1
                elif line.startswith(str(end1)) and not fix_end1:
                    end1 = end1 + 1
                else:
                    if line.startswith("Date Time"):
                        fix_end1 = True
                    try:  # 找到第一个首元素可以转为日期的行
                        cell0 = line.split("\t")[0]
                        if cell0:
                            _ = pd.Timestamp(cell0)
                            start2 = i
                            break
                    except ValueError:
                        continue

            data_title = read_csv_ex(file, sep="\t", skiprows=start1, nrows=(end1 - 1))
            data_value = read_csv_ex(file, sep="\t", skiprows=start2 - 1)
            data_value = data_value.applymap(lambda x: np.nan if str(x).strip() == "" else x)
            data_value.dropna(axis=1, how="all", inplace=True)  # 剔除空值
            name = ""
            for name in data_title.columns:  # 列标题中有空白字符，直接用"Description"取不到描述列
                if str(name).strip().startswith("Description"):
                    break
            if name == "":
                logger.warning(f"未找到{file}点名中的描述列，请检查")
                return None
            point_name = list(data_title[name])
            point_name = [p.strip() for p in point_name]  # 删除点名中的空白字符
            point_name.insert(0, "DateTime")
            data_value.set_axis(point_name, axis=1, inplace=True)
            data_value.to_csv(outfile1, index=False)
            return deal_dcs_pre_file(outfile1, time_dict)



import numpy as np


def func_transformer(func):
    '''
    transform this kind of function:
    ```
    def demo_func(x):
        x1, x2, x3 = x
        return x1 ** 2 + x2 ** 2 + x3 ** 2
    ```
    into this kind of function:
    ```
    def demo_func(x):
        x1, x2, x3 = x[:,0], x[:,1], x[:,2]
        return x1 ** 2 + (x2 - 0.05) ** 2 + x3 ** 2
    ```
    getting vectorial performance if possible
    :param func:
    :return:
    '''

    prefered_function_format = '''
    def demo_func(x):
        x1, x2, x3 = x[:, 0], x[:, 1], x[:, 2]
        return x1 ** 2 + (x2 - 0.05) ** 2 + x3 ** 2
    '''

    is_vector = getattr(func, 'is_vector', False)
    if is_vector:
        return func
    else:
        if func.__code__.co_argcount == 1:
            def func_transformed(X):
                return np.array([func(x) for x in X])

            return func_transformed
        elif func.__code__.co_argcount > 1:

            def func_transformed(X):
                return np.array([func(*tuple(x)) for x in X])

            return func_transformed

    raise ValueError('''
    object function error,
    function should be like this:
    ''' + prefered_function_format)



import numpy as np


def ranking(self):
    # GA select the biggest one, but we want to minimize func, so we put a negative here
    self.FitV = -self.Y


def ranking_linear(self):
    '''
    For more details see [Baker1985]_.

    :param self:
    :return:

    .. [Baker1985] Baker J E, "Adaptive selection methods for genetic
    algorithms, 1985.
    '''
    self.FitV = np.argsort(np.argsort(-self.Y))
    return self.FitV



# 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
import openai

openai.api_key = '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'

prompt = '你知道西安热工院吗？'


def completion(prompt):
    """
    根据提示续写文本内容

    :param prompt: 用户给出的提示文本
    :return:
    """
    engine = 'text-davinci-003'  # 两个模型
    engine = 'davinci'
    com = openai.Completion.create(engine=engine,
                                   prompt=prompt,
                                   max_tokens=1024,
                                   n=1,
                                   stop=None,
                                   temperature=0.5)

    response = com.choices[0].text
    return response


def chat(question):
    """
    聊天机器人，
    :param question: 用户发出的提问或回复
    :return:
    """
    model = 'gpt-3.5-turbo'
    model = 'gpt-3.5-turbo-0301'
    response = openai.ChatCompletion.create(
        model=model,
        messages=[
            {"role": "system", "content": "一个有10年Python开发经验的资深算法工程师"},  # 机器人的人设，
            {"role": "user", "content": question}  # 表示是用户提问内容
        ]
    )

    text = response.get("choices")[0]["message"]["content"]
    return text


print(chat(prompt))



import argparse
import torch
import torchvision
import os
from sys import platform
import shutil
from yangke.objDetect.models import Darknet
import cv2
import time
import numpy as np
from pathlib import Path
import glob
from threading import Thread
import random

img_formats = ['.bmp', '.jpg', '.jpeg', '.png', '.tif', '.dng']
vid_formats = ['.mov', '.avi', '.mp4']


class LoadImages:  # for inference
    def __init__(self, path, img_size=416):
        path = str(Path(path))  # os-agnostic
        files = []
        if os.path.isdir(path):
            files = sorted(glob.glob(os.path.join(path, '*.*')))
        elif os.path.isfile(path):
            files = [path]

        images = [x for x in files if os.path.splitext(x)[-1].lower() in img_formats]
        videos = [x for x in files if os.path.splitext(x)[-1].lower() in vid_formats]
        nI, nV = len(images), len(videos)

        self.img_size = img_size
        self.files = images + videos
        self.nF = nI + nV  # number of files
        self.video_flag = [False] * nI + [True] * nV
        self.mode = 'images'
        if any(videos):
            self.new_video(videos[0])  # new video
        else:
            self.cap = None
        assert self.nF > 0, 'No images or videos found in ' + path

    def __iter__(self):
        self.count = 0
        return self

    def __next__(self):
        if self.count == self.nF:
            raise StopIteration
        path = self.files[self.count]

        if self.video_flag[self.count]:
            # Read video
            self.mode = 'video'
            ret_val, img0 = self.cap.read()
            if not ret_val:
                self.count += 1
                self.cap.release()
                if self.count == self.nF:  # last video
                    raise StopIteration
                else:
                    path = self.files[self.count]
                    self.new_video(path)
                    ret_val, img0 = self.cap.read()

            self.frame += 1
            print('video %g/%g (%g/%g) %s: ' % (self.count + 1, self.nF, self.frame, self.nframes, path), end='')

        else:
            # Read image
            self.count += 1
            img0 = cv2.imread(path)  # BGR
            assert img0 is not None, 'Image Not Found ' + path
            print('image %g/%g %s: ' % (self.count, self.nF, path), end='')

        # Padded resize
        img = letterbox(img0, new_shape=self.img_size)[0]

        # Convert
        img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
        img = np.ascontiguousarray(img)

        # cv2.imwrite(path + '.letterbox.jpg', 255 * img.transpose((1, 2, 0))[:, :, ::-1])  # save letterbox image
        return path, img, img0, self.cap

    def new_video(self, path):
        self.frame = 0
        self.cap = cv2.VideoCapture(path)
        self.nframes = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

    def __len__(self):
        return self.nF  # number of files


class LoadStreams:  # multiple IP or RTSP cameras
    def __init__(self, sources='streams.txt', img_size=416):
        self.mode = 'images'
        self.img_size = img_size

        if os.path.isfile(sources):
            with open(sources, 'r') as f:
                sources = [x.strip() for x in f.read().splitlines() if len(x.strip())]
        else:
            sources = [sources]

        n = len(sources)
        self.imgs = [None] * n
        self.sources = sources
        for i, s in enumerate(sources):
            # Start the thread to read frames from the video stream
            print('%g/%g: %s... ' % (i + 1, n, s), end='')
            cap = cv2.VideoCapture(0 if s == '0' else s)
            assert cap.isOpened(), 'Failed to open %s' % s
            w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            fps = cap.get(cv2.CAP_PROP_FPS) % 100
            _, self.imgs[i] = cap.read()  # guarantee first frame
            thread = Thread(target=self.update, args=([i, cap]), daemon=True)
            print(' success (%gx%g at %.2f FPS).' % (w, h, fps))
            thread.start()
        print('')  # newline

        # check for common shapes
        s = np.stack([letterbox(x, new_shape=self.img_size)[0].shape for x in self.imgs], 0)  # inference shapes
        self.rect = np.unique(s, axis=0).shape[0] == 1  # rect inference if all shapes equal
        if not self.rect:
            print('WARNING: Different stream shapes detected. For optimal performance supply similarly-shaped streams.')

    def update(self, index, cap):
        # Read next stream frame in a daemon thread
        n = 0
        while cap.isOpened():
            n += 1
            # _, self.imgs[index] = cap.read()
            cap.grab()
            if n == 4:  # read every 4th frame
                _, self.imgs[index] = cap.retrieve()
                n = 0
            time.sleep(0.01)  # wait time

    def __iter__(self):
        self.count = -1
        return self

    def __next__(self):
        self.count += 1
        img0 = self.imgs.copy()
        if cv2.waitKey(1) == ord('q'):  # q to quit
            cv2.destroyAllWindows()
            raise StopIteration

        # Letterbox
        img = [letterbox(x, new_shape=self.img_size, auto=self.rect)[0] for x in img0]

        # Stack
        img = np.stack(img, 0)

        # Convert
        img = img[:, :, :, ::-1].transpose(0, 3, 1, 2)  # BGR to RGB, to bsx3x416x416
        img = np.ascontiguousarray(img)

        return self.sources, img, img0, None

    def __len__(self):
        return 0  # 1E12 frames = 32 streams at 30 FPS for 30 years


def letterbox(img, new_shape=(416, 416), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True):
    # Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
    shape = img.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = max(new_shape) / max(shape)
    if not scaleup:  # only scale down, do not scale up (for better test mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, 64), np.mod(dh, 64)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = new_shape
        ratio = new_shape[0] / shape[1], new_shape[1] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return img, ratio, (dw, dh)


def load_classes(path):
    # Loads *.names file at 'path'
    with open(path, 'r') as f:
        names = f.read().split('\n')
    return list(filter(None, names))  # filter removes empty strings (such as last line)


def select_device(device='', apex=False, batch_size=None):
    # device = 'cpu' or '0' or '0,1,2,3'
    device = str(device)
    cpu_request = device.lower() == 'cpu'
    if device and not cpu_request:  # if device requested other than 'cpu'
        os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
        assert torch.cuda.is_available(), 'CUDA unavailable, invalid device %s requested' % device  # check availablity

    cuda = False if cpu_request else torch.cuda.is_available()
    if cuda:
        c = 1024 ** 2  # bytes to MB
        ng = torch.cuda.device_count()
        if ng > 1 and batch_size:  # check that batch_size is compatible with device_count
            assert batch_size % ng == 0, 'batch-size %g not multiple of GPU count %g' % (batch_size, ng)
        x = [torch.cuda.get_device_properties(i) for i in range(ng)]
        s = 'Using CUDA ' + ('Apex ' if apex else '')  # apex for mixed precision https://github.com/NVIDIA/apex
        for i in range(0, ng):
            if i == 1:
                s = ' ' * len(s)
            print("%sdevice%g _CudaDeviceProperties(name='%s', total_memory=%dMB)" %
                  (s, i, x[i].NAME, x[i].total_memory / c))
    else:
        print('Using CPU')

    print('')  # skip a line
    return torch.device('cuda:0' if cuda else 'cpu')


def load_classifier(name='resnet101', n=2):
    # Loads a pretrained model reshaped to n-class output
    import pretrainedmodels  # https://github.com/Cadene/pretrained-models.pytorch#torchvision
    model = pretrainedmodels.__dict__[name](num_classes=1000, pretrained='imagenet')

    # Display model properties
    for x in ['model.input_size', 'model.input_space', 'model.input_range', 'model.mean', 'model.std']:
        print(x + ' =', eval(x))

    # Reshape output to n classes
    filters = model.last_linear.weight.shape[1]
    model.last_linear.bias = torch.nn.Parameter(torch.zeros(n))
    model.last_linear.weight = torch.nn.Parameter(torch.zeros(n, filters))
    model.last_linear.out_features = n
    return model


def box_iou(box1, box2):
    # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Arguments:
        box1 (Tensor[N, 4])
        box2 (Tensor[M, 4])
    Returns:
        iou (Tensor[N, M]): the NxM matrix containing the pairwise
            IoU values for every element in boxes1 and boxes2
    """

    def box_area(box):
        # box = 4xn
        return (box[2] - box[0]) * (box[3] - box[1])

    area1 = box_area(box1.t())
    area2 = box_area(box2.t())

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    inter = (torch.min(box1[:, None, 2:], box2[:, 2:]) - torch.max(box1[:, None, :2], box2[:, :2])).clamp(0).prod(2)
    return inter / (area1[:, None] + area2 - inter)  # iou = inter / (area1 + area2 - inter)


def time_synchronized():
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    return time.time()


def xyxy2xywh(x):
    """
    将 xyxy 形式的矩形框表达式转换为 xywh 形式，不进行归一化处理

    :param x: 形状为(:,4)的数组，4列分别对应矩形框的 左上角x坐标，左上角y坐标，右下角x坐标，右下角y坐标
    :return: 形状为(:,4)的数组，4列分别对应矩形框的 x_center，y_center, 宽, 高
    """
    # Transform box coordinates from [x1, y1, x2, y2] (where xy1=top-left, xy2=bottom-right) to [x, y, w, h]
    y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height
    return y


def xywh2xyxy(x):
    # Transform box coordinates from [x, y, w, h] to [x1, y1, x2, y2] (where xy1=top-left, xy2=bottom-right)
    y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def non_max_suppression(prediction, conf_thres=0.1, iou_thres=0.6, multi_label=True, classes=None, agnostic=False):
    """
    Performs  Non-Maximum Suppression on inference results
    Returns detections with shape:
        nx6 (x1, y1, x2, y2, conf, cls)
    """

    # Box constraints
    min_wh, max_wh = 2, 4096  # (pixels) minimum and maximum box width and height

    method = 'merge'
    nc = prediction[0].shape[1] - 5  # number of classes
    multi_label &= nc > 1  # multiple labels per box
    output = [None] * len(prediction)

    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply conf constraint
        x = x[x[:, 4] > conf_thres]

        # Apply width-height constraint
        x = x[((x[:, 2:4] > min_wh) & (x[:, 2:4] < max_wh)).all(1)]

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Compute conf
        x[..., 5:] *= x[..., 4:5]  # conf = obj_conf * cls_conf

        # Box (center x, center y, width, height) to (x1, y1, x2, y2)
        box = xywh2xyxy(x[:, :4])

        # Detections matrix nx6 (xyxy, conf, cls)
        if multi_label:
            i, j = (x[:, 5:] > conf_thres).nonzero().t()
            x = torch.cat((box[i], x[i, j + 5].unsqueeze(1), j.float().unsqueeze(1)), 1)
        else:  # best class only
            conf, j = x[:, 5:].max(1)
            x = torch.cat((box, conf.unsqueeze(1), j.float().unsqueeze(1)), 1)

        # Filter by class
        if classes:
            x = x[(j.view(-1, 1) == torch.tensor(classes, device=j.device)).any(1)]

        # Apply finite constraint
        if not torch.isfinite(x).all():
            x = x[torch.isfinite(x).all(1)]

        # If none remain process next image
        n = x.shape[0]  # number of boxes
        if not n:
            continue

        # Sort by confidence
        # if method == 'fast_batch':
        #    x = x[x[:, 4].argsort(descending=True)]

        # Batched NMS
        c = x[:, 5] * 0 if agnostic else x[:, 5]  # classes
        boxes, scores = x[:, :4].clone() + c.view(-1, 1) * max_wh, x[:, 4]  # boxes (offset by class), scores
        if method == 'merge':  # Merge NMS (boxes merged using weighted mean)
            i = torchvision.ops.boxes.nms(boxes, scores, iou_thres)
            if n < 1E4:  # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
                # weights = (box_iou(boxes, boxes).tril_() > iou_thres) * scores.view(-1, 1)  # box weights
                # weights /= weights.sum(0)  # normalize
                # x[:, :4] = torch.mm(weights.T, x[:, :4])
                weights = (box_iou(boxes[i], boxes) > iou_thres) * scores[None]  # box weights
                x[i, :4] = torch.mm(weights / weights.sum(1, keepdim=True), x[:, :4]).float()  # merged boxes
        elif method == 'vision':
            i = torchvision.ops.boxes.nms(boxes, scores, iou_thres)
        elif method == 'fast':  # FastNMS from https://github.com/dbolya/yolact
            iou = box_iou(boxes, boxes).triu_(diagonal=1)  # upper triangular iou matrix
            i = iou.max(0)[0] < iou_thres

        output[xi] = x[i]
    return output


def clip_coords(boxes, img_shape):
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    boxes[:, 0].clamp_(0, img_shape[1])  # x1
    boxes[:, 1].clamp_(0, img_shape[0])  # y1
    boxes[:, 2].clamp_(0, img_shape[1])  # x2
    boxes[:, 3].clamp_(0, img_shape[0])  # y2


def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None):
    """


    :param img1_shape:
    :param coords:
    :param img0_shape:
    :param ratio_pad:
    :return:
    """
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = max(img1_shape) / max(img0_shape)  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    coords[:, [0, 2]] -= pad[0]  # x padding
    coords[:, [1, 3]] -= pad[1]  # y padding
    coords[:, :4] /= gain
    clip_coords(coords, img0_shape)
    return coords


def apply_classifier(x, model, img, im0):
    # applies a second stage classifier to yolo outputs
    im0 = [im0] if isinstance(im0, np.ndarray) else im0
    for i, d in enumerate(x):  # per image
        if d is not None and len(d):
            d = d.clone()

            # Reshape and pad cutouts
            b = xyxy2xywh(d[:, :4])  # boxes
            b[:, 2:] = b[:, 2:].max(1)[0].unsqueeze(1)  # rectangle to square
            b[:, 2:] = b[:, 2:] * 1.3 + 30  # pad
            d[:, :4] = xywh2xyxy(b).long()

            # Rescale boxes from img_size to im0 size
            scale_coords(img.shape[2:], d[:, :4], im0[i].shape)

            # Classes
            pred_cls1 = d[:, 5].long()
            ims = []
            for j, a in enumerate(d):  # per item
                cutout = im0[i][int(a[1]):int(a[3]), int(a[0]):int(a[2])]
                im = cv2.resize(cutout, (224, 224))  # BGR
                # cv2.imwrite('test%i.jpg' % j, cutout)

                im = im[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
                im = np.ascontiguousarray(im, dtype=np.float32)  # uint8 to float32
                im /= 255.0  # 0 - 255 to 0.0 - 1.0
                ims.append(im)

            pred_cls2 = model(torch.Tensor(ims).to(d.device)).argmax(1)  # classifier prediction
            x[i] = x[i][pred_cls1 == pred_cls2]  # retain matching class detections

    return x


# Plotting functions ---------------------------------------------------------------------------------------------------
def plot_one_box(x, img, color=None, label=None, line_thickness=None):
    # Plots one bounding box on image img
    tl = line_thickness or round(0.002 * (img.shape[0] + img.shape[1]) / 2) + 1  # line/font thickness
    color = color or [random.randint(0, 255) for _ in range(3)]
    c1, c2 = (int(x[0]), int(x[1])), (int(x[2]), int(x[3]))
    cv2.rectangle(img, c1, c2, color, thickness=tl)
    if label:
        tf = max(tl - 1, 1)  # font thickness
        t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
        c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
        cv2.rectangle(img, c1, c2, color, -1)  # filled
        cv2.putText(img, label, (c1[0], c1[1] - 2), 0, tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)


def detect(save_img=False):
    img_size = opt.img_size  # (320, 192) or (416, 256) or (608, 352) for (height, width)
    out, source, weights, half, view_img, save_txt = opt.output, opt.source, opt.weights, opt.half, opt.view_img, opt.save_txt
    webcam = source == '0' or source.startswith('rtsp') or source.startswith('http') or source.endswith('.txt')

    # Initialize
    device = select_device(opt.device)
    if os.path.exists(out):
        shutil.rmtree(out)  # delete output folder
    os.makedirs(out)  # make new output folder

    # Initialize model
    model: Darknet = Darknet(opt.cfg, img_size)

    # Load weights
    if os.path.exists(weights):
        if weights.endswith('.pt'):  # pytorch format
            model.load_state_dict(torch.load(weights, map_location=device)['model'], strict=False)
    else:
        print("模型文件不存在，请检查！ {}".format(os.path.abspath(weights)))
        exit(0)

    # Second-stage classifier
    classify = False
    if classify:
        modelc = load_classifier(name='resnet34', n=2)  # initialize
        modelc.load_state_dict(torch.load('weights/resnet101.pt', map_location=device)['model'])  # load weights
        modelc.to(device).eval()

    # Eval mode
    model.to(device).eval()

    # Fuse Conv2d + BatchNorm2d layers
    # model.fuse()

    # Half precision
    half = half and device.type != 'cpu'  # half precision only supported on CUDA
    if half:
        model.half()

    # Set Dataloader
    vid_path, vid_writer = None, None
    if webcam:
        view_img = True
        torch.backends.cudnn.benchmark = True  # set True to speed up constant image size inference
        dataset = LoadStreams(source, img_size=img_size)
    else:
        save_img = True
        dataset = LoadImages(source, img_size=img_size)

    # Get names and colors
    names = load_classes(opt.names)
    colors = [[random.randint(0, 255) for _ in range(3)] for _ in range(len(names))]

    # Run inference
    t0 = time.time()
    _ = model(torch.zeros((1, 3, img_size, img_size), device=device)) if device.type != 'cpu' else None  # run once
    for path, img, im0s, vid_cap in dataset:
        img = torch.from_numpy(img).to(device)
        img = img.half() if half else img.float()  # uint8 to fp16/32
        img /= 255.0  # 0 - 255 to 0.0 - 1.0
        if img.ndimension() == 3:
            img = img.unsqueeze(0)

        # Inference
        t1 = time_synchronized()
        pred = model(img, augment=opt.augment)[0]
        t2 = time_synchronized()

        # to float
        if half:
            pred = pred.float()

        # Apply NMS
        pred = non_max_suppression(pred, opt.conf_thres, opt.iou_thres,
                                   multi_label=False, classes=opt.classes, agnostic=opt.agnostic_nms)

        # Apply Classifier
        if classify:
            pred = apply_classifier(pred, modelc, img, im0s)

        # Process detections
        for i, det in enumerate(pred):  # detections per image
            if webcam:  # batch_size >= 1
                p, s, im0 = path[i], '%g: ' % i, im0s[i]
            else:
                p, s, im0 = path, '', im0s

            save_path = str(Path(out) / Path(p).name)
            s += '%gx%g ' % img.shape[2:]  # print string
            if det is not None and len(det):
                # Rescale boxes from img_size to im0 size
                det[:, :4] = scale_coords(img.shape[2:], det[:, :4], im0.shape).round()

                # Print results
                for c in det[:, -1].unique():
                    n = (det[:, -1] == c).sum()  # detections per class
                    s += '%g %ss, ' % (n, names[int(c)])  # add to string

                # Write results
                for *xyxy, conf, cls in det:
                    if save_txt:  # Write to file
                        with open(save_path + '.txt', 'a') as file:
                            file.write(('%g ' * 6 + '\n') % (*xyxy, cls, conf))

                    if save_img or view_img:  # Add bbox to image
                        label = '%s %.2f' % (names[int(cls)], conf)
                        plot_one_box(xyxy, im0, label=label, color=colors[int(cls)])

            # Print time (inference + NMS)
            print('%sDone. (%.3fs)' % (s, t2 - t1))

            # Stream results
            if view_img:
                cv2.imshow(p, im0)
                if cv2.waitKey(1) == ord('q'):  # q to quit
                    raise StopIteration

            # Save results (image with detections)
            if save_img:
                if dataset.mode == 'images':
                    cv2.imwrite(save_path, im0)
                else:
                    if vid_path != save_path:  # new video
                        vid_path = save_path
                        if isinstance(vid_writer, cv2.VideoWriter):
                            vid_writer.release()  # release previous video writer

                        fps = vid_cap.get(cv2.CAP_PROP_FPS)
                        w = int(vid_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                        h = int(vid_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                        vid_writer = cv2.VideoWriter(save_path, cv2.VideoWriter_fourcc(*opt.fourcc), fps, (w, h))
                    vid_writer.write(im0)

    if save_txt or save_img:
        print('Results saved to %s' % os.getcwd() + os.sep + out)
        if platform == 'darwin':  # MacOS
            os.system('open ' + out + ' ' + save_path)

    print('Done. (%.3fs)' % (time.time() - t0))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--cfg', type=str, default='cfg/yolov3-tiny.cfg', help='*.cfg path')
    parser.add_argument('--names', type=str, default='data/coco.names', help='*.names path')
    parser.add_argument('--weights', type=str, default='D:\\Users\\pycharm\\face\\yolov3-tiny.pt', help='weights path')
    parser.add_argument('--source', type=str, default=r'D:\Users\pycharm\face\image',
                        help='source')  # input file/folder, 0 for webcam
    parser.add_argument('--output', type=str, default='output', help='output folder')  # output folder
    parser.add_argument('--img-size', type=int, default=512, help='inference size (pixels)')
    parser.add_argument('--conf-thres', type=float, default=0.3, help='object confidence threshold')
    parser.add_argument('--iou-thres', type=float, default=0.6, help='IOU threshold for NMS')
    parser.add_argument('--fourcc', type=str, default='mp4v', help='output video codec (verify ffmpeg support)')
    parser.add_argument('--half', action='store_true', help='half precision FP16 inference')
    parser.add_argument('--device', default='', help='device id (i.e. 0 or 0,1) or cpu')
    parser.add_argument('--view-img', action='store_true', help='display results')
    parser.add_argument('--save-txt', action='store_true', help='save results to *.txt')
    parser.add_argument('--classes', nargs='+', type=int, help='filter by class')
    parser.add_argument('--agnostic-nms', action='store_true', help='class-agnostic NMS')
    parser.add_argument('--augment', action='store_true', help='augmented inference')
    opt = parser.parse_args()
    print(opt)
    with torch.no_grad():
        detect()



import os
import shutil

from pathlib2 import Path

from yangke.base import search_file_in_folders
from yangke.core import runCMD
from yangke.common.config import logger

spider_paths = [Path(__file__).parent.parent / Path("spider"), Path(__file__).parent.parent, os.getcwd()]


def start_scrapy_start(start_crawl_file: str, charset="utf8"):
    """
    开启爬虫项目，使用start.py文件

    :param start_file: 爬虫项目中的开始文件，如r"D:\stock10jqka\start_stock10jqka.py"
    :return:
    """
    # 实现方式，在spider_file目录运行cmd命令
    # python -m scrapy runspider myspider.py
    # 实现方式，在start_file目录运行cmd命令
    # python start_spider.py
    logger.info(start_crawl_file)
    if Path(start_crawl_file).is_absolute():
        pass
    else:
        base_name = Path(start_crawl_file).name
        start_crawl_file = search_file_in_folders(base_name, spider_paths)
    logger.info(f"爬虫启动文件：{start_crawl_file}")
    cwd = Path(start_crawl_file).parent
    logger.info(f"爬虫运行目录：{cwd}")

    # 把settings.yaml复制到爬虫的运行目录，以使爬虫能使用os.getcwd()中配置文件的配置
    settings_file_origin = os.path.join(os.getcwd(), "settings.yaml")
    settings_file_origin1 = os.path.join(os.getcwd(), "settings.yml")
    settings_file_dst = os.path.join(str(cwd), "settings.yaml")
    if os.path.exists(settings_file_origin):
        shutil.copy2(settings_file_origin, settings_file_dst)
    elif os.path.exists(settings_file_origin1):
        shutil.copy2(settings_file_origin1, settings_file_dst)

    runCMD('python "{}"'.format(start_crawl_file), charset=charset, wait_for_result=True, cwd=cwd,
           output_type="REALTIME_NORETURN")


def start_scrapy_spider(spider_file: str, charset="utf8"):
    """
    开启指定爬虫爬取网页

    :param spider_file: 爬虫项目中的爬虫文件，如r"D:\stock10jqka\stock10jqka\spiders\jqka_spider.py"，如果是相对路径，默认会在yangke/spider目录下查找
    :return:
    """
    if Path(spider_file).is_absolute():
        pass
    else:
        base_name = Path(spider_file).name
        spider_file = search_file_in_folders(base_name, spider_paths)
    cwd = Path(spider_file).parent
    runCMD('python -m scrapy runspider "{}"'.format(spider_file), charset=charset, wait_for_result=True, cwd=cwd,
           output_type="REALTIME_NORETURN")



import argparse
import json

from torch.utils.data import DataLoader


from .models import *
from .yolo_train import *


def coco80_to_coco91_class():  # converts 80-index (val2014) to 91-index (paper)
    # https://tech.amikelive.com/node-718/what-object-categories-labels-are-in-coco-dataset/
    # a = np.loadtxt('data/coco.names', dtype='str', delimiter='\n')
    # b = np.loadtxt('data/coco_paper.names', dtype='str', delimiter='\n')
    # x1 = [list(a[i] == b).index(True) + 1 for i in range(80)]  # darknet to coco
    # x2 = [list(b[i] == a).index(True) if any(b[i] == a) else None for i in range(91)]  # coco to darknet
    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 27, 28, 31, 32, 33, 34,
         35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
         64, 65, 67, 70, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 84, 85, 86, 87, 88, 89, 90]
    return x


def time_synchronized():
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    return time.time()


def compute_ap(recall, precision):
    """ Compute the average precision, given the recall and precision curves.
    Source: https://github.com/rbgirshick/py-faster-rcnn.
    # Arguments
        recall:    The recall curve (list).
        precision: The precision curve (list).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """

    # Append sentinel values to beginning and end
    mrec = np.concatenate(([0.], recall, [min(recall[-1] + 1E-3, 1.)]))
    mpre = np.concatenate(([0.], precision, [0.]))

    # Compute the precision envelope
    mpre = np.flip(np.maximum.accumulate(np.flip(mpre)))

    # Integrate area under curve
    method = 'interp'  # methods: 'continuous', 'interp'
    if method == 'interp':
        x = np.linspace(0, 1, 101)  # 101-point interp (COCO)
        ap = np.trapz(np.interp(x, mrec, mpre), x)  # integrate
    else:  # 'continuous'
        i = np.where(mrec[1:] != mrec[:-1])[0]  # points where x axis (recall) changes
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])  # area under curve

    return ap


def ap_per_class(tp, conf, pred_cls, target_cls):
    """ Compute the average precision, given the recall and precision curves.
    Source: https://github.com/rafaelpadilla/Object-Detection-Metrics.
    # Arguments
        tp:    True positives (nparray, nx1 or nx10).
        conf:  Objectness value from 0-1 (nparray).
        pred_cls: Predicted object classes (nparray).
        target_cls: True object classes (nparray).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """

    # Sort by objectness
    i = np.argsort(-conf)
    tp, conf, pred_cls = tp[i], conf[i], pred_cls[i]

    # Find unique classes
    unique_classes = np.unique(target_cls)

    # Create Precision-Recall curve and compute AP for each class
    pr_score = 0.1  # score to evaluate P and R https://github.com/ultralytics/yolov3/issues/898
    s = [len(unique_classes), tp.shape[1]]  # number class, number iou thresholds (i.e. 10 for mAP0.5...0.95)
    ap, p, r = np.zeros(s), np.zeros(s), np.zeros(s)
    for ci, c in enumerate(unique_classes):
        i = pred_cls == c
        n_gt = (target_cls == c).sum()  # Number of ground truth objects
        n_p = i.sum()  # Number of predicted objects

        if n_p == 0 or n_gt == 0:
            continue
        else:
            # Accumulate FPs and TPs
            fpc = (1 - tp[i]).cumsum(0)
            tpc = tp[i].cumsum(0)

            # Recall
            recall = tpc / (n_gt + 1e-16)  # recall curve
            r[ci] = np.interp(-pr_score, -conf[i], recall[:, 0])  # r at pr_score, negative x, xp because xp decreases

            # Precision
            precision = tpc / (tpc + fpc)  # precision curve
            p[ci] = np.interp(-pr_score, -conf[i], precision[:, 0])  # p at pr_score

            # AP from recall-precision curve
            for j in range(tp.shape[1]):
                ap[ci, j] = compute_ap(recall[:, j], precision[:, j])

            # Plot
            # fig, ax = plt.subplots(1, 1, figsize=(5, 5))
            # ax.plot(recall, precision)
            # ax.set_xlabel('Recall')
            # ax.set_ylabel('Precision')
            # ax.set_xlim(0, 1.01)
            # ax.set_ylim(0, 1.01)
            # fig.tight_layout()
            # fig.savefig('PR_curve.png', dpi=300)

    # Compute F1 score (harmonic mean of precision and recall)
    f1 = 2 * p * r / (p + r + 1e-16)

    return p, r, ap, f1, unique_classes.astype('int32')


def test(cfg,
         data,
         weights=None,
         batch_size=16,
         img_size=416,
         conf_thres=0.001,
         iou_thres=0.6,  # for nms
         save_json=False,
         single_cls=False,
         augment=False,
         model=None,
         dataloader=None):
    # Initialize/load model and set device
    if model is None:
        device = select_device(opt.device, batch_size=batch_size)
        verbose = opt.task == 'test'

        # Remove previous
        for f in glob.glob('test_batch*.png'):
            os.remove(f)

        # Initialize model
        model = Darknet(cfg, img_size)

        # Load weights
        attempt_download(weights)

        model.load_state_dict(torch.load(weights, map_location=device)['model'])

        # Fuse
        model.fuse()
        model.to(device)

        if device.type != 'cpu' and torch.cuda.device_count() > 1:
            model = nn.DataParallel(model)
    else:  # called by train.py
        device = next(model.parameters()).device  # get model device
        verbose = False

    # Configure run
    # data = parse_data_cfg(data)
    data = {  # data_dict = parse_data_cfg(data)
        "classes": get_settings('aiTrain.darknet.class'),
        "names": get_settings('aiTrain.darknet.names'),
        "train": get_settings('aiTrain.darknet.train'),
        "valid": get_settings('aiTrain.darknet.valid'),
        "trainLabelFolder": get_settings('aiTrain.darknet.trainLabelFolder') or None,  # 替换默认Label文件中的路径，保留图片名不变
        "validLabelFolder": get_settings('aiTrain.darknet.validLabelFolder') or None,
    }
    nc = 1 if single_cls else int(data['classes'])  # number of classes
    path = data['valid']  # path to test images
    names = load_classes(data['names'])  # class names
    iouv = torch.linspace(start=0.5, end=0.95, steps=10).to(device)  # iou vector for mAP@0.5:0.95，steps是步数，不是步长
    iouv = iouv[0].view(1)  # comment for mAP@0.5:0.95
    niou = iouv.numel()

    # Dataloader
    if dataloader is None:  # 如果测试数据集的Dataloader
        dataset = DarknetDataset(path, label_folder=get_settings('aiTrain.darknet.validLabelFolder'), img_size=img_size,
                                 batch_size=batch_size, rect=True, single_cls=opt.single_cls)
        batch_size = min(batch_size, len(dataset))
        dataloader = DataLoader(dataset,
                                batch_size=batch_size,
                                num_workers=min([os.cpu_count(), batch_size if batch_size > 1 else 0, 8]),
                                pin_memory=True,
                                collate_fn=dataset.collate_fn)

    seen = 0
    model.eval()
    # 如果不是cpu训练，就是用批大小为1的单张图片作为输入，是用model(input)进行测试计算
    _ = model(torch.zeros((1, 3, img_size, img_size), device=device)) if device.type != 'cpu' else None  # run once
    coco91class = coco80_to_coco91_class()
    s = ('%20s' + '%10s' * 6) % ('Class', 'Images', 'Targets', 'P', 'R', 'mAP@0.5', 'F1')
    p, r, f1, mp, mr, map, mf1, t0, t1 = 0., 0., 0., 0., 0., 0., 0., 0., 0.
    loss = torch.zeros(3, device=device)
    jdict, stats, ap, ap_class = [], [], [], []
    for batch_i, (imgs, targets, paths, shapes) in enumerate(tqdm(dataloader, desc=s)):
        imgs = imgs.to(device).float() / 255.0  # uint8 to float32, 0 - 255 to 0.0 - 1.0
        targets = targets.to(device)
        nb, _, height, width = imgs.shape  # batch size, channels, height, width
        whwh = torch.Tensor([width, height, width, height]).to(device)

        # Plot images with bounding boxes
        f = 'test_batch%g.png' % batch_i  # filename
        if batch_i < 1 and not os.path.exists(f):
            plot_images(imgs=imgs, targets=targets, paths=paths, fname=f)

        # Disable gradients
        with torch.no_grad():
            # Run model
            t = time_synchronized()
            inf_out, train_out = model(imgs, augment=augment)  # inference and training outputs
            t0 += time_synchronized() - t

            # Compute loss
            if hasattr(model, 'hyp'):  # if model has loss hyperparameters
                loss += compute_loss(train_out, targets, model)[1][:3]  # GIoU, obj, cls

            # Run NMS
            t = time_synchronized()
            output = non_max_suppression(inf_out, conf_thres=conf_thres, iou_thres=iou_thres)  # nms
            t1 += time_synchronized() - t

        # Statistics per image
        for si, pred in enumerate(output):
            labels = targets[targets[:, 0] == si, 1:]
            nl = len(labels)
            tcls = labels[:, 0].tolist() if nl else []  # target class
            seen += 1

            if pred is None:
                if nl:
                    stats.append((torch.zeros(0, niou, dtype=torch.bool), torch.Tensor(), torch.Tensor(), tcls))
                continue

            # Append to text file
            # with open('test.txt', 'a') as file:
            #    [file.write('%11.5g' * 7 % tuple(x) + '\n') for x in pred]

            # Clip boxes to image bounds
            clip_coords(pred, (height, width))

            # Append to pycocotools JSON dictionary
            if save_json:
                # [{"image_id": 42, "category_id": 18, "bbox": [258.15, 41.29, 348.26, 243.78], "score": 0.236}, ...
                image_id = int(Path(paths[si]).stem.split('_')[-1])
                box = pred[:, :4].clone()  # xyxy
                scale_coords(imgs[si].shape[1:], box, shapes[si][0], shapes[si][1])  # to original shape
                box = xyxy2xywh(box)  # xywh
                box[:, :2] -= box[:, 2:] / 2  # xy center to top-left corner
                for p, b in zip(pred.tolist(), box.tolist()):
                    jdict.append({'image_id': image_id,
                                  'category_id': coco91class[int(p[5])],
                                  'bbox': [round(x, 3) for x in b],
                                  'score': round(p[4], 5)})

            # Assign all predictions as incorrect
            correct = torch.zeros(pred.shape[0], niou, dtype=torch.bool, device=device)
            if nl:
                detected = []  # target indices
                tcls_tensor = labels[:, 0]

                # target boxes
                tbox = xywh2xyxy(labels[:, 1:5]) * whwh

                # Per target class
                for cls in torch.unique(tcls_tensor):
                    ti = (cls == tcls_tensor).nonzero().view(-1)  # prediction indices
                    pi = (cls == pred[:, 5]).nonzero().view(-1)  # target indices

                    # Search for detections
                    if pi.shape[0]:
                        # Prediction to target ious
                        ious, i = box_iou(pred[pi, :4], tbox[ti]).max(1)  # best ious, indices

                        # Append detections
                        for j in (ious > iouv[0]).nonzero():
                            d = ti[i[j]]  # detected target
                            if d not in detected:
                                detected.append(d)
                                correct[pi[j]] = ious[j] > iouv  # iou_thres is 1xn
                                if len(detected) == nl:  # all targets already located in image
                                    break

            # Append statistics (correct, conf, pcls, tcls)
            stats.append((correct.cpu(), pred[:, 4].cpu(), pred[:, 5].cpu(), tcls))

    # Compute statistics
    stats = [np.concatenate(x, 0) for x in zip(*stats)]  # to numpy
    if len(stats):
        p, r, ap, f1, ap_class = ap_per_class(*stats)
        if niou > 1:
            p, r, ap, f1 = p[:, 0], r[:, 0], ap.mean(1), ap[:, 0]  # [P, R, AP@0.5:0.95, AP@0.5]
        mp, mr, map, mf1 = p.mean(), r.mean(), ap.mean(), f1.mean()
        nt = np.bincount(stats[3].astype(np.int64), minlength=nc)  # number of targets per class
    else:
        nt = torch.zeros(1)

    # Print results
    pf = '%20s' + '%10.3g' * 6  # print format
    print(pf % ('all', seen, nt.sum(), mp, mr, map, mf1))

    # Print results per class
    if verbose and nc > 1 and len(stats):
        for i, c in enumerate(ap_class):
            print(pf % (names[c], seen, nt[c], p[i], r[i], ap[i], f1[i]))

    # Print speeds
    if verbose or save_json:
        t = tuple(x / seen * 1E3 for x in (t0, t1, t0 + t1)) + (img_size, img_size, batch_size)  # tuple
        print('Speed: %.1f/%.1f/%.1f ms inference/NMS/total per %gx%g image at batch-size %g' % t)

    # Save JSON
    if save_json and map and len(jdict):
        print('\nCOCO mAP with pycocotools...')
        imgIds = [int(Path(x).stem.split('_')[-1]) for x in dataloader.dataset.img_files]
        with open('results.json', 'w') as file:
            json.dump(jdict, file)

        try:
            from pycocotools.coco import COCO
            from pycocotools.cocoeval import COCOeval
        except:
            print('WARNING: missing pycocotools package, can not compute official COCO mAP. See requirements.txt.')

        # https://github.com/cocodataset/cocoapi/blob/master/PythonAPI/pycocoEvalDemo.ipynb
        cocoGt = COCO(glob.glob(r'D:\Users\pycharm\face\data\annotations/instances_val*.json')[0])  # initialize COCO ground truth api
        cocoDt = cocoGt.loadRes('results.json')  # initialize COCO pred api

        cocoEval = COCOeval(cocoGt, cocoDt, 'bbox')
        cocoEval.params.imgIds = imgIds  # [:32]  # only evaluate these images
        cocoEval.evaluate()
        cocoEval.accumulate()
        cocoEval.summarize()
        mf1, map = cocoEval.stats[:2]  # update to pycocotools results (mAP@0.5:0.95, mAP@0.5)

    # Return results
    maps = np.zeros(nc) + map
    for i, c in enumerate(ap_class):
        maps[c] = ap[i]
    return (mp, mr, map, mf1, *(loss.cpu() / len(dataloader)).tolist()), maps


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='win_x64.py')
    parser.add_argument('--cfg', type=str, default='cfg/yolov3-spp.cfg', help='*.cfg path')
    parser.add_argument('--data', type=str, default='data/coco2014.data', help='*.data path')
    parser.add_argument('--weights', type=str, default='weights/yolov3-spp-ultralytics.pt', help='weights path')
    parser.add_argument('--batch-size', type=int, default=16, help='size of each image batch')
    parser.add_argument('--img-size', type=int, default=416, help='inference size (pixels)')
    parser.add_argument('--conf-thres', type=float, default=0.001, help='object confidence threshold')
    parser.add_argument('--iou-thres', type=float, default=0.6, help='IOU threshold for NMS')
    parser.add_argument('--save-json', action='store_true', help='save a cocoapi-compatible JSON results file')
    parser.add_argument('--task', default='test', help="'test', 'study', 'benchmark'")
    parser.add_argument('--device', default='', help='device id (i.e. 0 or 0,1) or cpu')
    parser.add_argument('--single-cls', action='store_true', help='train as single-class dataset')
    parser.add_argument('--augment', action='store_true', help='augmented inference')
    opt = parser.parse_args()
    opt.save_json = opt.save_json or any([x in opt.data for x in ['coco.data', 'coco2014.data', 'coco2017.data']])
    print(opt)

    # task = 'test', 'study', 'benchmark'
    if opt.task == 'test':  # (default) test normally
        test(opt.cfg,
             opt.data,
             opt.weights,
             opt.batch_size,
             opt.img_size,
             opt.conf_thres,
             opt.iou_thres,
             opt.save_json,
             opt.single_cls,
             opt.augment)

    elif opt.task == 'benchmark':  # mAPs at 320-608 at conf 0.5 and 0.7
        y = []
        for i in [320, 416, 512, 608]:  # img-size
            for j in [0.5, 0.7]:  # iou-thres
                t = time.time()
                r = test(opt.cfg, opt.data, opt.weights, opt.batch_size, i, opt.conf_thres, j, opt.save_json)[0]
                y.append(r + (time.time() - t,))
        np.savetxt('benchmark.txt', y, fmt='%10.4g')  # y = np.loadtxt('study.txt')

    elif opt.task == 'study':  # Parameter study
        y = []
        x = np.arange(0.4, 0.9, 0.05)  # iou-thres
        for i in x:
            t = time.time()
            r = test(opt.cfg, opt.data, opt.weights, opt.batch_size, opt.img_size, opt.conf_thres, i, opt.save_json)[0]
            y.append(r + (time.time() - t,))
        np.savetxt('study.txt', y, fmt='%10.4g')  # y = np.loadtxt('study.txt')

        # Plot
        fig, ax = plt.subplots(3, 1, figsize=(6, 6))
        y = np.stack(y, 0)
        ax[0].plot(x, y[:, 2], marker='.', label='mAP@0.5')
        ax[0].set_ylabel('mAP')
        ax[1].plot(x, y[:, 3], marker='.', label='mAP@0.5:0.95')
        ax[1].set_ylabel('mAP')
        ax[2].plot(x, y[:, -1], marker='.', label='time')
        ax[2].set_ylabel('time (s)')
        for i in range(3):
            ax[i].legend()
            ax[i].set_xlabel('iou_thr')
        fig.tight_layout()
        plt.savefig('study.jpg', dpi=200)



suffix_p_01: "去标题行" = "_p预处理_修正"
# noinspection PyUnresolvedReferences
suffix_p_10: "剪切时间" = "_剪切p时间_修正"
# noinspection PyUnresolvedReferences
suffix_p_20: "汇总" = "p汇总_修正"
# noinspection PyUnresolvedReferences
suffix_dcs_01 = "_dcs预处理_修正"
# noinspection PyUnresolvedReferences
suffix_dcs_10: "剪切时间" = "_剪切dcs时间_修正"
# noinspection PyUnresolvedReferences
suffix_dcs_20: "汇总" = "dcs汇总_修正"
# noinspection PyUnresolvedReferences
suffix_imp_10: "剪切时间" = "_剪切imp时间_修正"
# noinspection PyUnresolvedReferences
suffix_imp_11 = "_大气压修正"
# noinspection PyUnresolvedReferences
suffix_imp_20: "汇总" = "imp汇总_修正"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2019/8/20
# @Author  : github.com/guofei9987

import numpy as np
from yangke.sko.tools import func_transformer
from .base import SkoBase


class PSO(SkoBase):
    """
    Do PSO (Particle swarm optimization) algorithm.

    This algorithm was adapted from the earlier works of J. Kennedy and
    R.C. Eberhart in Particle Swarm Optimization [IJCNN1995]_.

    The position update can be defined as:

    .. math::

       x_{i}(t+1) = x_{i}(t) + v_{i}(t+1)

    Where the position at the current step :math:`t` is updated using
    the computed velocity at :math:`t+1`. Furthermore, the velocity update
    is defined as:

    .. math::

       v_{ij}(t + 1) = w * v_{ij}(t) + c_{p}r_{1j}(t)[y_{ij}(t) − x_{ij}(t)]
                       + c_{g}r_{2j}(t)[\hat{y}_{j}(t) − x_{ij}(t)]

    Here, :math:`cp` and :math:`cg` are the cognitive and social parameters
    respectively. They control the particle's behavior given two choices: (1) to
    follow its *personal best* or (2) follow the swarm's *global best* position.
    Overall, this dictates if the swarm is explorative or exploitative in nature.
    In addition, a parameter :math:`w` controls the inertia of the swarm's
    movement.

    .. [IJCNN1995] J. Kennedy and R.C. Eberhart, "Particle Swarm Optimization,"
    Proceedings of the IEEE International Joint Conference on Neural
    Networks, 1995, pp. 1942-1948.

    Parameters
    --------------------
    func : function
        The func you want to do optimal
    dim : int
        Number of dimension, which is number of parameters of func.
    pop : int
        Size of population, which is the number of Particles. We use 'pop' to keep accordance with GA
    max_iter : int
        Max of iter iterations

    Attributes
    ----------------------
    pbest_x : array_like, shape is (pop,dim)
        best location of every particle in history
    pbest_y : array_like, shape is (pop,1)
        best image of every particle in history
    gbest_x : array_like, shape is (1,dim)
        general best location for all particles in history
    gbest_y : float
        general best image  for all particles in history
    gbest_y_hist : list
        gbest_y of every iteration


    Examples
    -----------------------------
    see https://scikit-opt.github.io/scikit-opt/#/en/README?id=_3-psoparticle-swarm-optimization
    """

    def __init__(self, func, dim, pop=40, max_iter=150, lb=None, ub=None, w=0.8, c1=0.5, c2=0.5):
        self.func = func_transformer(func)
        self.w = w  # inertia
        self.cp, self.cg = c1, c2  # parameters to control personal best, global best respectively
        self.pop = pop  # number of particles
        self.dim = dim  # dimension of particles, which is the number of variables of func
        self.max_iter = max_iter  # max iter

        self.has_constraints = not (lb is None and ub is None)
        self.lb = -np.ones(self.dim) if lb is None else np.array(lb)
        self.ub = np.ones(self.dim) if ub is None else np.array(ub)
        assert self.dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.dim))
        v_high = self.ub - self.lb
        self.V = np.random.uniform(low=-v_high, high=v_high, size=(self.pop, self.dim))  # speed of particles
        self.Y = self.cal_y()  # y = f(x) for all particles
        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = self.Y.copy()  # best image of every particle in history
        self.gbest_x = np.zeros((1, self.dim))  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

        # record verbose values
        self.record_mode = False
        self.record_value = {'X': [], 'V': [], 'Y': []}
        self.best_x, self.best_y = self.gbest_x, self.gbest_y  # history reasons, will be deprecated

    def update_V(self):
        r1 = np.random.rand(self.pop, self.dim)
        r2 = np.random.rand(self.pop, self.dim)
        self.V = self.w * self.V + \
                 self.cp * r1 * (self.pbest_x - self.X) + \
                 self.cg * r2 * (self.gbest_x - self.X)

    def update_X(self):
        self.X = self.X + self.V

        if self.has_constraints:
            self.X = np.clip(self.X, self.lb, self.ub)

    def cal_y(self):
        # calculate y for every x in X
        self.Y = self.func(self.X).reshape(-1, 1)
        return self.Y

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        self.pbest_x = np.where(self.pbest_y > self.Y, self.X, self.pbest_x)
        self.pbest_y = np.where(self.pbest_y > self.Y, self.Y, self.pbest_y)

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        if self.gbest_y > self.Y.min():
            self.gbest_x = self.X[self.Y.argmin(), :].copy()
            self.gbest_y = self.Y.min()

    def recorder(self):
        if not self.record_mode:
            return
        self.record_value['X'].append(self.X)
        self.record_value['V'].append(self.V)
        self.record_value['Y'].append(self.Y)

    def run(self, max_iter=None):
        self.max_iter = max_iter or self.max_iter
        for iter_num in range(self.max_iter):
            self.update_V()
            self.recorder()
            self.update_X()
            self.cal_y()
            self.update_pbest()
            self.update_gbest()

            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

    fit = run



import torch


def mutation(self):
    '''
    mutation of 0/1 type chromosome
    faster than `self.Chrom = (mask + self.Chrom) % 2`
    :param self:
    :return:
    '''
    mask = (torch.rand(size=(self.size_pop, self.len_chrom), device=self.device) < self.prob_mut).type(torch.int8)
    self.Chrom ^= mask
    return self.Chrom



import math
from functools import partial

from yangke.common.config import logger
from yangke.sis.RuiJin.tags import TagsRead, TagsWrite, TagsReadDCS
from yangke.sis.cold_optimize_base import Condenser
from yangke.sis.dll_file import init_write_sis, init_dbp_api, read_data, get_tag_value, ColdResult, 标煤热值, \
    read_data_from_dcs
from yangke.base import execute_function_by_interval
from yangke.performance.iapws97 import get_h_by_px, get_h_by_pt, get_h_by_tx, get_t_by_ph, get_p_by_t, get_t_by_p

available_type = [(2, 2, 2), (2, 1, 2), (2, 2, 0), (0, 2, 2), (2, 0, 2), (2, 1, 0), (0, 1, 2)]
cold_result = ColdResult()
condenser = Condenser()

condenser.set_info(类型='双背压', 设计凝汽器热负荷=494.68 + 490.692, 有效长度=14, 管子外径=25, 管子壁厚=0.5,
                   t_cyc_in=22.3, bp=4.75, 有效换热面积=64000, 洁净度=0.85,
                   循环水体积流量=101000, 设计循环水流速=2.131
                   )


def get_power_cold(on_a3, on_b3, on_c3):
    """
    计算循环水泵耗功

    AC开，则功率为2000+2100=4100kW 4200
    ABC开，则功率为2200+2350+2250=6800

    :param on_a3:
    :param on_b3:
    :param on_c3:
    :return:
    """
    res = 0
    if on_a3 == 2 and on_b3 == 2 and on_c3 == 2:  # 三高
        res = 6800
    elif on_a3 == 2 and on_b3 == 2 and on_c3 == 0:  # 两高
        res = 4200
    elif on_a3 == 0 and on_b3 == 2 and on_c3 == 2:  # 两高
        res = 4200
    elif on_a3 == 2 and on_b3 == 0 and on_c3 == 2:  # 两高
        res = 4200  # ...  # 4100，B泵高速耗功比A、C泵高
    elif on_a3 == 2 and on_b3 == 1 and on_c3 == 2:  # 两高一低
        res = 2200 + 1650 + 2200
    elif on_a3 == 2 and on_b3 == 1 and on_c3 == 0:  # 一高一低
        res = 2200 + 1600
    elif on_a3 == 0 and on_b3 == 1 and on_c3 == 2:  # 一高一低
        res = 2200 + 1600
    return res


def get_flow(on_a3, on_b3, on_c3):
    """
    获取每种循泵运行方式下的循环水流量

    :param on_a3:
    :param on_b3:
    :param on_c3:
    :return:
    """
    flow_cycle = 0
    if on_a3 == 2 and on_b3 == 2 and on_c3 == 2:  # 三高
        flow_cycle = 101051
    elif on_a3 == 2 and on_b3 == 2 and on_c3 == 0:  # 两高
        flow_cycle = 79478
    elif on_a3 == 0 and on_b3 == 2 and on_c3 == 2:  # 两高
        flow_cycle = 79478
    elif on_a3 == 2 and on_b3 == 0 and on_c3 == 2:  # 两高
        flow_cycle = 79478
    elif on_a3 == 2 and on_b3 == 1 and on_c3 == 2:  # 两高一低
        flow_cycle = 93764
    elif on_a3 == 2 and on_b3 == 1 and on_c3 == 0:  # 一高一低
        flow_cycle = 69246  # AB低：69246，
    elif on_a3 == 0 and on_b3 == 1 and on_c3 == 2:  # 一高一低
        flow_cycle = 69246
    return flow_cycle


def get_power_by_bp(bp, bp_iter, power):
    """
    获取背压变化后，机组的功率，使用设计的背压曲线

    :param bp:
    :param bp_iter:
    :param power:
    :return:
    """
    x = bp
    percent = 0.005457 * x ** 4 - 0.166231 * x ** 3 + 1.893091 * x ** 2 - 8.552442 * x + 12.976042
    percent = 1 - percent / 100
    x = bp_iter
    percent_iter = 0.005457 * x ** 4 - 0.166231 * x ** 3 + 1.893091 * x ** 2 - 8.552442 * x + 12.976042
    percent_iter = 1 - percent_iter / 100
    power_init = power / percent
    power_iter = power_init * percent_iter
    return power_iter


def get_coal_by_bp(bp, bp_iter, coal):
    """
    获取背压变化后，机组的热耗或煤耗，使用设计的背压曲线，煤耗和热耗的变化率是相同的，因此可以使用同一条曲线

    :param
    bp:
    :param
    bp_iter:
    :param
    power:
    :return:
    """
    x = bp
    percent = -0.005199 * x ** 4 + 0.160356 * x ** 3 - 1.840626 * x ** 2 + 8.346903 * x - 12.678426
    percent = 1 - percent / 100
    x = bp_iter
    percent_iter = -0.005199 * x ** 4 + 0.160356 * x ** 3 - 1.840626 * x ** 2 + 8.346903 * x - 12.678426
    percent_iter = 1 - percent_iter / 100
    coal_init = coal / percent
    coal_iter = coal_init * percent_iter
    return coal_iter


def output_res(dis_res):
    """
    返回计算结果至前端画面

    :param dis_res:
    :return:
    """
    pass


def optimize_coal(snapshot, group_num=3):
    """
    以煤耗为目标对循环水泵运行方式进行优化；优化假设条件：1、优化前后凝汽器运行补水流量不变；

    :param snapshot:
    :param group_num: 机组编号
    :return:
    """
    opt_result = {}
    get_value = partial(get_tag_value, snapshot)
    if group_num == 3:
        power = get_value(TagsRead.power3, 903.28)
        t_env = get_value(TagsRead.环境温度, 14.55)
        f_condense = get_value(TagsRead.凝泵出口凝结水流量3, 1555.46)
        f_makeup = get_value(TagsRead.凝汽器运行补水流量3, 20.13)
        bp1 = get_value(TagsRead.bph_a_3, 4.33)  # 高压凝汽器压力
        bp2 = get_value(TagsRead.bph_b_3, 4.33)  # 高压凝汽器压力
        bp3 = get_value(TagsRead.bph_c_3, 4.33)  # 高压凝汽器压力
        bp4 = get_value(TagsRead.bpl_a_3, 4.33)  # 低压凝汽器压力
        bp5 = get_value(TagsRead.bpl_b_3, 4.33)  # 低压凝汽器压力
        bp6 = get_value(TagsRead.bpl_c_3, 4.33)  # 低压凝汽器压力
        t_well = get_value(TagsRead.热井出水温度3, 30.88)  # 按出水温度计算焓
        t_makeup = get_value(TagsRead.凝汽器补水温度3, 30.92)
        t_cyc_cold1_3 = get_value(TagsRead.凝汽器1进口A循环水温度1_3, 15.75)
        t_cyc_cold2_3 = get_value(TagsRead.凝汽器1进口A循环水温度2_3, 15.91)
        t_cyc_cold3_3 = get_value(TagsRead.凝汽器1进口B循环水温度1_3, 15.85)
        t_cyc_cold4_3 = get_value(TagsRead.凝汽器1进口B循环水温度2_3, 15.74)
        t_cyc_hot1 = get_value(TagsRead.凝汽器2出口A循环水温度1_3, 27.38)
        t_cyc_hot2 = get_value(TagsRead.凝汽器2出口A循环水温度2_3, 27.42)
        t_cyc_hot3 = get_value(TagsRead.凝汽器2出口B循环水温度1_3, 27.69)
        t_cyc_hot4 = get_value(TagsRead.凝汽器2出口B循环水温度2_3, 27.54)
    else:
        power = get_value(TagsRead.power4, 903.28)
        t_env = get_value(TagsRead.环境温度, 14.55)
        f_condense = get_value(TagsRead.凝泵出口凝结水流量4, 1555.46)
        f_makeup = get_value(TagsRead.凝汽器运行补水流量4, 20.13)
        bp1 = get_value(TagsRead.bph_a_4, 4.33)  # 高压凝汽器压力
        bp2 = get_value(TagsRead.bph_b_4, 4.33)  # 高压凝汽器压力
        bp3 = get_value(TagsRead.bph_c_4, 4.33)  # 高压凝汽器压力
        bp4 = get_value(TagsRead.bpl_a_4, 4.33)  # 低压凝汽器压力
        bp5 = get_value(TagsRead.bpl_b_4, 4.33)  # 低压凝汽器压力
        bp6 = get_value(TagsRead.bpl_c_4, 4.33)  # 低压凝汽器压力
        t_well = get_value(TagsRead.热井出水温度4, 30.88)  # 按出水温度计算焓
        t_makeup = get_value(TagsRead.凝汽器补水温度4, 30.92)
        t_cyc_cold1_3 = get_value(TagsRead.凝汽器1进口A循环水温度1_4, 15.75)
        t_cyc_cold2_3 = get_value(TagsRead.凝汽器1进口A循环水温度2_4, 15.91)
        t_cyc_cold3_3 = get_value(TagsRead.凝汽器1进口B循环水温度1_4, 15.85)
        t_cyc_cold4_3 = get_value(TagsRead.凝汽器1进口B循环水温度2_4, 15.74)
        t_cyc_hot1 = get_value(TagsRead.凝汽器2出口A循环水温度1_4, 27.38)
        t_cyc_hot2 = get_value(TagsRead.凝汽器2出口A循环水温度2_4, 27.42)
        t_cyc_hot3 = get_value(TagsRead.凝汽器2出口B循环水温度1_4, 27.69)
        t_cyc_hot4 = get_value(TagsRead.凝汽器2出口B循环水温度2_4, 27.54)

    if power < 10:  # 停机状态
        opt_result.update({"3": "stop"})
    else:
        f_out_lp = f_condense - f_makeup  # 0.7583 * f_condense3 + 32.22  # 低压缸排汽流量
        heat_dispatch = 3484097.04 * power + 333917622.65  # 近似认为凝汽器排汽能量只与功率相关

        bp = (bp1 + bp2 + bp3 + bp4 + bp5 + bp6) / 6
        del bp1, bp2, bp3, bp4, bp5, bp6
        # h_well3 = get_h_by_px(bp / 1000, 0)  # 凝汽器压力对应的饱和水的焓
        h_well = get_h_by_tx(t_well, 0)
        h_makeup = get_h_by_pt(2.28, t_makeup)

        # MJ/h = 3.6 * kW
        heat_load = heat_dispatch / 1000 - f_out_lp * h_well - f_makeup * h_makeup  # 热负荷=排汽能量-凝结水能量-补水
        heat_load_kW = heat_load / 3.6  # kW
        del t_makeup, h_makeup, f_makeup

        t_cyc_cold = (t_cyc_cold1_3 + t_cyc_cold2_3 + t_cyc_cold3_3 + t_cyc_cold4_3) / 4
        t_cyc_hot = (t_cyc_hot1 + t_cyc_hot2 + t_cyc_hot3 + t_cyc_hot4) / 4
        del t_cyc_cold1_3, t_cyc_cold2_3, t_cyc_cold3_3, t_cyc_cold4_3, t_cyc_hot1, t_cyc_hot2, t_cyc_hot3, t_cyc_hot4
        h_cyc_cold = get_h_by_pt(0.3, t_cyc_cold)
        h_cyc_hot = get_h_by_pt(0.24, t_cyc_hot)

        t_sat_bp = get_t_by_p(bp / 1000)  # 凝汽器压力对应的饱和温度
        凝汽器端差 = t_sat_bp - t_cyc_hot
        凝汽器过冷度 = t_sat_bp - t_well
        热井出水温度与循环水进水温度差 = t_well - t_cyc_cold
        delta_t_cyc = t_cyc_hot - t_cyc_cold  # 循环水温升
        delta_h_cyc = h_cyc_hot - h_cyc_cold  # 循环水焓升
        环境温度与冷却塔出水温度差 = t_cyc_cold - t_env
        实际循环水流量 = heat_load / (h_cyc_hot - h_cyc_cold)

        rpm_a3 = get_value(TagsRead.循泵A转速3, 372.85)
        rpm_b3 = get_value(TagsRead.循泵B转速3, 330.6)
        rpm_c3 = get_value(TagsRead.循泵C转速3, 0)
        on_a3 = 2 if rpm_a3 > 350 else 0
        on_b3 = 1 if rpm_b3 > 300 else 0
        on_b3 = 2 if rpm_b3 > 350 else on_b3
        on_c3 = 2 if rpm_c3 > 350 else 0
        del rpm_a3, rpm_b3, rpm_c3

        power_cold = get_power_cold(on_a3, on_b3, on_c3)  # 当前运行方式下的功率
        flow_cycle = get_flow(on_a3, on_b3, on_c3)

        coal_rate_roll10 = get_value(TagsRead.入炉煤量)  # t/h，取10min的平均值
        power_roll10 = get_value(TagsRead.power3)
        if power_roll10 <= 100:
            logger.debug(f"机组处于停机或启动状态，无法计算煤耗")
            当前煤耗 = 0
        else:
            当前煤耗 = 1000 * coal_rate_roll10 / power_roll10

        # 调试
        当前煤耗 = 287  # g/kWh
        coal_exclude_pump = 当前煤耗 * power / (power - power_cold / 1000)  # 扣除冷端耗功后的发电煤耗

        price_coal = get_value(TagsRead.price_coal, 1200)
        price_power = get_value(TagsRead.price_power, 0.5)
        benefit = coal_exclude_pump / 1000 / 1000 * price_coal * (power - power_cold) * 1000  # g/kWh->t/kWh->元/kWh->元/h

        cond_list = []
        for a, b, c in available_type:
            if (a, b, c) == (on_a3, on_b3, on_c3) or (c, b, a) == (on_a3, on_b3, on_c3):
                flow_cycle_iter = 实际循环水流量
            else:
                flow_cycle_iter = get_flow(a, b, c)

            power_cold_iter = get_power_cold(a, b, c)

            # ---------------------------- 计算当前运行方式下的背压 ------------------------------------
            # 认为循环水流量增大对凝汽器和冷却塔换热性能的影响是相同的
            delta_h_iter = heat_load / flow_cycle_iter
            h_cyc_cold_iter = h_cyc_cold + (delta_h_cyc - delta_h_iter) * 1.5 / 3
            h_cyc_hot_iter = h_cyc_cold_iter + delta_h_iter
            t_cyc_cold_iter = get_t_by_ph(0.3, h_cyc_cold_iter)  # 当前运行方式下循环水冷侧温度
            t_cyc_hot_iter = get_t_by_ph(0.24, h_cyc_hot_iter)
            t_sat_bp_iter = t_sat_bp + (t_cyc_hot_iter - t_cyc_hot)  # 认为循环水温度变化=热井水温度变化=凝汽器饱和温度变化

            bp_iter = get_p_by_t(t_sat_bp_iter) * 1000  # kPa
            # ---------------------------- 计算当前运行方式下的背压 ------------------------------------

            # ---------------------------- 计算背压变化导致的功率变化 -----------------------------------
            delta_bp_iter = bp_iter - bp
            power_iter = get_power_by_bp(bp, bp_iter, power)
            coal_iter = get_coal_by_bp(bp, bp_iter, 当前煤耗)  # 当前迭代步的十分钟平均煤耗
            coal_exclude_pump_iter = coal_iter * power_iter / (power_iter - power_cold_iter / 1000)  # 扣除冷端耗功后的煤耗
            delta_power_gen = (power_iter - power) * 1000  # kW,发电机多发的电量
            delta_power_cold = power_cold_iter - power_cold  # kW，循泵多耗的电量
            delta_power_net = delta_power_gen - delta_power_cold  # 净功率变化
            benefit_iter = coal_exclude_pump_iter / 1000 * price_coal * (power_iter - power_cold_iter)  # 元/h

            cond_list.append({"循泵运行方式": [a, b, c],
                              "净功率": power_iter - power_cold_iter / 1000,
                              "净功率增加": delta_power_net,
                              "发电机功率": power_iter,
                              "发电机功率增加": delta_power_gen,
                              "冷端耗功": power_cold_iter,
                              "冷端耗功增加": delta_power_cold,
                              "煤耗": coal_exclude_pump_iter,
                              "煤耗降低": coal_exclude_pump - coal_exclude_pump_iter,
                              "背压": bp_iter,
                              "发电收益": benefit_iter,
                              })
            # ---------------------------- 计算背压变化导致的功率变化 -----------------------------------

        obj = get_tag_value(snapshot, TagsRead.优化目标, 0)  # 0表示煤耗，1表示收益
        if obj == 0:
            cond_list.sort(key=lambda x: x["净功率增加"])
        else:
            cond_list.sort(key=lambda x: x["发电收益"])

        # 写出优化结果
        res = cond_list[0]
        if group_num == 3:
            dis_res = {
                TagsWrite.循泵A启停3: res["循泵运行方式"][0],
                TagsWrite.循泵B启停3: res["循泵运行方式"][1],
                TagsWrite.循泵C启停3: res["循泵运行方式"][2],
                TagsWrite.凝汽器热负荷3: heat_load,
                TagsWrite.凝汽器端差3: 凝汽器端差,
                TagsWrite.凝汽器过冷度3: 凝汽器过冷度,
                TagsWrite.循环水流量3: 实际循环水流量,
                TagsWrite.冷端耗功3: power_cold,
                TagsWrite.冷端耗功优化值3: res["冷端耗功"],
                TagsWrite.发电功率优化值3: res["发电机功率"],
                TagsWrite.净功率3: power - power_cold / 1000,
                TagsWrite.净功率优化值3: res["净功率"],
                TagsWrite.煤耗3: coal_exclude_pump,  # 扣除冷端耗功的发电煤耗
                TagsWrite.煤耗优化值3: res["煤耗"],
                TagsWrite.发电收益3: benefit,
                TagsWrite.发电收益优化值3: res["发电收益"],
            }
        else:
            dis_res = {
                TagsWrite.循泵A启停4: res["循泵运行方式"][0],
                TagsWrite.循泵B启停4: res["循泵运行方式"][1],
                TagsWrite.循泵C启停4: res["循泵运行方式"][2],
                TagsWrite.凝汽器热负荷4: heat_load,
                TagsWrite.凝汽器端差4: 凝汽器端差,
                TagsWrite.凝汽器过冷度4: 凝汽器过冷度,
                TagsWrite.循环水流量4: 实际循环水流量,
                TagsWrite.冷端耗功4: power_cold,
                TagsWrite.冷端耗功优化值4: res["冷端耗功"],
                TagsWrite.发电功率优化值4: res["发电机功率"],
                TagsWrite.净功率4: power - power_cold / 1000,
                TagsWrite.净功率优化值4: res["净功率"],
                TagsWrite.煤耗4: coal_exclude_pump,  # 扣除冷端耗功的发电煤耗
                TagsWrite.煤耗优化值4: res["煤耗"],
                TagsWrite.发电收益4: benefit,
                TagsWrite.发电收益优化值4: res["发电收益"],
            }

        db_res = {"环境温度": t_env, "凝汽器热负荷": heat_load, "循泵运行方式": res["循泵运行方式"],
                  "循环水冷侧温度": t_cyc_cold}
        output_res(dis_res)


def optimize():
    snapshot = read_data(TagsRead, ip="172.18.240.191", port='12084', user="admin", password="admin")  # 未知原因读不到数据
    snapshot = read_data_from_dcs(TagsReadDCS, url='http://172.18.248.80:8080/nodiot/restful/redisValue')
    optimize_coal(snapshot, 3)
    optimize_coal(snapshot, 4)


init_write_sis(ip="172.18.240.191", port='12084', user="admin", passwd_str="admin")
execute_function_by_interval(optimize, minute=0, second=10)



"""
爬虫的调试：
使用pycharm以新项目模式打开spider下爬虫对应的文件夹即可
    例如调试stock10jqka爬虫，pycharm操作为：【File】->【Open...】->选择*/yangke/spider/stock10jqka项目->【Ok】
"""
import stock10jqka.start_stock10jqka
import stock10jqka.stock10jqka.spiders.jqka_spider



"""
data_analysis.py的数据处理方法
data_analysis.py负责维护GUI界面，这里放计算方法
"""
from yangke.performance.basic_para import Case, alternative_symbol
import pandas as pd
from yangke.core import str_in_list
from yangke.base import extend_list_by_alternative_symbols
import copy
from yangke.common.config import logger


def update_data(imp_file, p_file=None, dcs_file=None):
    """
    将试验测点名称对应的数据赋值给计算程序中相应参数

    :param df:
    :return:
    """
    cases = {}
    df = pd.read_excel(imp_file, index_col=0)
    values = df.values
    case_names = list(df.columns)  # 计算工况名称，每项对应一个工况
    var_names = list(df.index)
    current_case = Case()  # 当前工况，临时变量
    # 将各参数归类到Case类对应的变量上
    annotations = Case.__annotations__
    case_para_loc = {}  # 记录计算程序参数对应的试验测点在df数据结构中的位置
    for name in annotations.keys():  # 计算程序中参数名称
        case_para_loc.update({name: []})  # 每个计算程序参数可能对应多个试验测点，计算中取平均值，这里将多个测点位置记录在列表中
    # -------------------------- 首先确定试验测点对应哪个计算程序参数 ---------------------------------
    for i, name in enumerate(var_names):  # 试验测点的名称
        flag = False
        for k, v in annotations.items():  # 计算程序参数k，和其可能的名称v
            name_sub_str_list = v.split(",")  # 将可能的名称v转为列表
            name_sub_str_list = extend_list_by_alternative_symbols(name_sub_str_list, alternative_symbol)
            if str_in_list(name, name_sub_str_list, revert=True):  # 如果计算程序参数k可能的名称命中试验参数名称，则记录
                case_para_loc.get(k).append(i)
                flag = True
                break  # 如果当前试验测点已经对应一个计算程序参数，则不可能对应另一个计算程序参数，因此跳出当前循环
        if flag is False:  # 说明没有找到测点对应的计算程序中的参数名
            logger.warning(f"测点名称({name})对应的计算程序参数名未知！")

    logger.debug(case_para_loc)
    for i, case_name in enumerate(case_names):  # 遍历所有工况
        for k, loc in case_para_loc.items():  # 遍历所有测点
            if len(loc) != 0:
                # 将测点名和测点数值传给TestPoint对象，让其初始化测点数据
                temp_series = df[case_name].iloc[loc]
                name_list = list(temp_series.index)
                value_list = temp_series.values
                current_case.set_value_of_test_point(var_str=k, test_point_name_list=name_list,
                                                     value_list=value_list)
        cases.update({case_name: copy.deepcopy(current_case)})
    return cases


def generate_case_skeleton(case):
    """
    根据单个工况的数据，猜测生成机组的热力系统结构。

    :param case:
    :return:
    """
    logger.debug("生成热力系统结构")
    pass



# 可以使用yk.fun()直接调用的方法写在这里
# __package__ = "yangke"
import sys
import subprocess
import os
from yangke.common import fileOperate as fileOp
from yangke.common.config import logger


def str_in_list(string: str, str_list: list, revert=False, need_item=False):
    """
    判断list的某个子元素是否包含指定字符串str

    用以和runCMD()配合使用，runCMD返回的的运行结果是list
    output, err=runCMD(...)
    strInList("result str", output)，如果output的某行包含"result str"，则返回True，否则返回False

    :param need_item: 是否需要返回找到的list中的项，默认不返回，设置为True则返回
    :param string: 字符串
    :param str_list: 字符串列表
    :param revert: 是否翻转，如果为True，则依次判断列表中的字符串是否在当前字符串中，为False则判断当前字符串是否在列表中的某个字符串中
    :return: bool/ (bool, str)
    """
    for line in str_list:
        if revert:
            if line in string:
                if need_item:
                    return True, line
                else:
                    return True
        else:
            if string in line:
                if need_item:
                    return True, line
                else:
                    return True
    if need_item:
        return False, None
    else:
        return False


def runCMD(command: str, charset: str = "utf8", wait_for_result: bool = True, cwd=None,
           output_type: str = "RETURN", timeout=None):
    """
    执行cmd命令，并返回标准输出结果，在windows系统下需要管理员权限运行的命令不能使用该方法调用。\n
    在linux系统下也可以使用。

    输出模式：支持"RETURN"(默认),"REALTIME_NORETURN"和"REALTIME_RETURN"，分别表示子进程的输出结果以字符串
    (output, err)形式返回、实时输出结果并且不返回、实时输出且需要以字符串形式返回。第三种方式的实时输出有短暂的缓冲期，实时性不如第二种
    方式，且第三种方式必须等待子线程运行结束

    :param timeout: 超时时间
    :param output_type: 输出模式，
    :param cwd: 工作目录
    :param command: 命令行命令
    :param charset: 编码方式
    :param wait_for_result: 是否等待命令执行完毕
    :return: 如果waitForResult=True，则返回执行结果字符串列表，否则，返回标准输出的接口
    """
    if "|" not in command:
        if output_type == "REALTIME_NORETURN":
            p = subprocess.Popen(command, shell=True, cwd=cwd)
            if wait_for_result:
                p.wait(timeout=timeout)  # 不重定向stdout和stderr，则p.wait()不会死锁
            return
        elif output_type == "RETURN":
            popen = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
            if wait_for_result:
                line, err = popen.communicate()  # command有大量输出时，popen.wait()可能会死锁
                # 这里获得的lines是字节码格式，pycharm显示为 b'CONTAINER ID ...\n' ，前面的b指的是bytes
                # 解码lines为字符串
                line = line.decode(charset)
                err = err.decode(charset)
                return line, err
            else:
                return popen.stdout, popen.stderr
        elif output_type == "REALTIME_RETURN":
            p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd)
            lines = []
            errs = []
            while p.poll() is None:
                """
                poll函数返回码：0-正常结束 1-sleep 2-子进程不存在 -15-kill None-在运行
                """
                if p.stdout is not None:
                    line = p.stdout.readline()
                    line = line.strip().decode(charset)
                    if line:
                        lines.append(line)
                        print(line)
                if p.stderr is not None:
                    err = p.stderr.readline()
                    err = err.strip().decode(charset)
                    if err:
                        errs.append(err)
                        print(err)
            return lines, errs
    else:  # python不支持直接运行command1 | findstr "targetStr" 这种cmd命令
        command = command.split("|")
        results, err = runCMD(command[0].strip())
        if command[1].strip().lower().startswith("findstr"):
            # 获得查找的目标字符串
            str_to_find = command[1].replace("findstr ", "")
            str_to_find = str_to_find.replace('"', "").strip()
            lines = []
            for line in results:
                if str_to_find.lower() in line.lower():
                    lines.append(line)
            return lines, err


def _runAsAdmin(cmd, cwd=None, charset="gbk", python=None, popup=True):
    """
    私有方法，管理员运行命令请使用 runAsAdmin()方法
    不能用于linux系统
    """
    import yangke.common.Interact as act
    # 因为引入包的关系，这个方法总是会被执行两次
    config = "Interact_temp_file"
    cwd = cwd or os.getcwd()  # os.path.dirname(__file__)  # 这里赋值调用目录，不能是文件目录
    config = os.path.join(cwd, config)
    if act.times == 1:  # 因为引入包的关系，这个方法总是会被执行两次，这里对真正的执行次数进行判断
        act.times = 0  # 如果执行过一次了，就直接返回，不再执行，且将执行次数恢复为0
        return
    result = act.runAdmin(cmd=cmd, cwd=cwd, charset=charset, python=python, popup=popup)  # runAdmin必须位于独立的文件中
    act.times = 1  # 将执行次数设置为1
    if os.path.exists(config):
        os.remove(config)

    return result


def runAsAdmin(cmd, cwd=None, charset="gbk", python=None, popup=True):
    """
    windows系统下以管理员身份运行命令

    :param cmd: 需要运行的命令
    :param cwd: 工作目录
    :param charset: 字符编码
    :param python: 需要运行的python解释器路径，默认为系统环境变脸的python解释器，如果系统环境没有配置，请手动指定
    :param popup: 是否弹出确认窗口，默认需要；windows安全策略不允许静默方式运行管理员命令，如果设置为False，则利用了windows的漏洞，在windows10系统会被windows defender当成病毒拦截，测试360安全卫士不会拦截，可以正常运行
    :return:
    """
    _runAsAdmin(cmd, cwd, charset, python, popup)  # 因为_runAsAdmin在第二次执行时返回None，因此只能从runResult文件中获得执行结果
    config = "Interact_temp_file"
    run_result = config + "_runResult"
    cwd = cwd or os.getcwd()
    run_result = os.path.join(cwd, run_result)
    result = fileOp.read_from_pickle(run_result)
    os.remove(run_result)  # 删除用于输出的临时文件
    return result


def exist_module(module_name: str):
    """
    判断指定的第三方库是否安装
    :return: bool
    """
    python_code = "import " + module_name
    try:
        exec(python_code)
    except ImportError:
        return False
    return True


def enable_docker():
    """
    判断docker是否在运行
    :return: Docker在运行返回True，否则返回False
    """
    title, err = runCMD('docker container ps')
    if len(title) == 0:
        logger.debug(err)
        print("Docker is not running, start Docker Desktop first!")
        return False
    else:
        return True


def status_of_container(container_name: str) -> str:
    """
    查询指定名称容器的状态，如果不存在指定名称的容器，则返回 None
    :param container_name:
    :return:
    """
    results, _ = runCMD('docker container ps -a')
    results = results.split("\n")
    for line in results:
        fields = list(filter(None, line.split("  ")))  # 以多个空格分隔line
        fields = [field.strip() for field in fields]  # 去除每个field中的空格
        if container_name in fields:
            return fields[4]
    return None



class FaceQualityCompleteness:
    def __init__(self, eyebrow: int, eye: int, nose: int, cheek: int, mouth: int, chin: int):
        """
        五官遮挡分，评价眉毛（Eyebrow）、眼睛（Eye）、鼻子（Nose）、脸颊（Cheek）、嘴巴（Mouth）、下巴（Chin）的被遮挡程度。

        :param eyebrow: 眉毛的遮挡分数[0,100]，分数越高遮挡越少。[0,80]表示发生遮挡。
        :param eye: 眼睛的遮挡分数[0,100],分数越高遮挡越少。[0,80]表示发生遮挡。
        :param nose: 鼻子的遮挡分数[0,100],分数越高遮挡越少。[0,60]表示发生遮挡。
        :param cheek: 脸颊的遮挡分数[0,100],分数越高遮挡越少。[0,70]表示发生遮挡。
        :param mouth: 嘴巴的遮挡分数[0,100],分数越高遮挡越少。[0,50]表示发生遮挡。
        :param chin: 下巴的遮挡分数[0,100],分数越高遮挡越少。[0,70]表示发生遮挡。
        """
        self.Eyebrow = eyebrow
        self.Eye = eye
        self.Nose = nose
        self.Cheek = cheek
        self.Mouth = mouth
        self.Chin = chin


class FaceQualityInfo:
    def __init__(self, score: int, sharpness: int, brightness: int, completeness: FaceQualityCompleteness):
        """
        人脸质量信息

        :param score: 质量分: [0,100]，综合评价图像质量是否适合人脸识别，分数越高质量越好。正常情况，只需要使用Score作为质量分总体的判断标准即可。Sharpness、Brightness、Completeness等细项分仅供参考。参考范围：[0,40]较差，[40,60] 一般，[60,80]较好，[80,100]很好。
        :param sharpness: 清晰分：[0,100]，评价图片清晰程度，分数越高越清晰。参考范围：[0,40]特别模糊，[40,60]模糊，[60,80]一般，[80,100]清晰。建议：人脸入库选取80以上的图片。
        :param brightness: 光照分：[0,100]，评价图片光照程度，分数越高越亮。参考范围： [0,30]偏暗，[30,70]光照正常，[70,100]偏亮。建议：人脸入库选取[30,70]的图片。
        :param completeness: 五官遮挡分，评价眉毛（Eyebrow）、眼睛（Eye）、鼻子（Nose）、脸颊（Cheek）、嘴巴（Mouth）、下巴（Chin）的被遮挡程度。
        """
        self.Score = score  # 可以由dlib库获得，detector.run(img, 1, -1)
        self.Sharpness = sharpness
        self.Brightness = brightness
        self.Completeness = completeness


class FaceHairAttributesInfo:
    def __init__(self, length: int, bang: int, color: int):
        """
        头发属性信息

        :param length: 0：光头，1：短发，2：中发，3：长发，4：绑发  注意：此字段可能返回 null，表示取不到有效值。
        :param bang: 0：有刘海，1：无刘海  注意：此字段可能返回 null，表示取不到有效值。
        :param color: 0：黑色，1：金色，2：棕色，3：灰白色  注意：此字段可能返回 null，表示取不到有效值。
        """
        self.Length = length
        self.Bang = bang
        self.Color = color


class FaceAttributesInfo:
    def __init__(self, gender: int, age: int, expression: int, glass: bool, pitch: int, yaw: int, roll: int,
                 beauty: int, hat: bool, mask: bool, hair: FaceHairAttributesInfo, eyeOpen: bool):
        """

        :param gender: 性别[0~49]为女性，[50，100]为男性，越接近0和100表示置信度越高。
        :param age: 年龄 [0~100]。NeedFaceAttributes 不为1或检测超过 5 张人脸时，此参数仍返回，但不具备参考意义。
        :param expression: 微笑[0(normal，正常)~50(smile，微笑)~100(laugh，大笑)]。NeedFaceAttributes 不为1 ，此参数仍返回，但不具备参考意义。
        :param glass: 是否有眼镜 [true,false]。
        :param pitch: 上下偏移[-30,30]，单位角度。
        :param yaw: 左右偏移[-30,30]，单位角度。
        :param roll: 平面旋转[-180,180]，单位角度。
        :param beauty: 魅力[0~100]。
        :param hat: 是否有帽子 [true,false]
        :param mask: 是否有口罩 [true,false]
        :param hair: 头发信息，包含头发长度（length）、有无刘海（bang）、头发颜色（color）
        :param eyeOpen: 双眼是否睁开 [true,false]
        """
        self.Hair = hair
        self.Mask = mask
        self.Hat = hat
        self.Beauty = beauty
        self.Roll = roll
        self.Yaw = yaw
        self.Pitch = pitch
        self.Glass = glass
        self.Expression = expression
        self.Gender = gender
        self.Age = age
        self.EyeOpen = eyeOpen


class FaceInfo:
    def __init__(self, x: int, y: int, width: int, height: int, faceAttributesInfo: FaceAttributesInfo,
                 faceQualityInfo: FaceQualityInfo):
        """
        人脸信息列表

        :param x: 人脸框左上角横坐标。
        :param y: 人脸框左上角纵坐标。
        :param width: 人脸框宽度
        :param height: 人脸框高度
        :param faceAttributesInfo: 人脸属性信息
        :param faceQualityInfo: 人脸质量信息
        """
        self.X = x
        self.Y = y
        self.Width = width
        self.Height = height
        self.FaceAttributesInfo = faceAttributesInfo
        self.FaceQualityInfo = faceQualityInfo



from direct.showbase.ShowBase import ShowBase
from math import pi, sin, cos
from direct.task import Task
from direct.actor.Actor import Actor, TransparencyAttrib
from direct.interval.IntervalGlobal import Sequence
from panda3d.core import Point3, ConfigVariableString, DirectionalLight, AmbientLight, TextNode
from panda3d.core import loadPrcFileData
from direct.gui.DirectGui import *

loadPrcFileData("", "#fullscreen true")


class MyApp(ShowBase):
    def __init__(self):
        ShowBase.__init__(self)
        self.scene = self.loader.loadModel("models/environment")
        self.scene.reparentTo(self.render)
        # 其他的场景根节点  self.render2d self.aspect2d self.pixel2d （self.hidden弃用）
        self.scene.setScale(0.25)
        self.scene.setPos(-8, 42, 0)

        # self.taskMgr.add(self.spinCameraTask, "SpinCameraTask")

        self.pandaActor = Actor("models/panda-model", {"walk": "models/panda-walk4"})
        self.pandaActor.setScale(0.005)
        # self.pandaActor.setTransparency(TransparencyAttrib.MAlpha)
        # self.pandaActor.setColorScale(1, 1, 1, 0.5)
        # self.pandaActor.setAlphaScale(0.5)
        self.pandaActor.reparentTo(self.render)
        self.pandaActor.loop("walk")
        # self.pandaActor.place()
        posInterval1 = self.pandaActor.posInterval(13,
                                                   Point3(0, -10, 0),
                                                   startPos=Point3(0, 10, 0))
        posInterval2 = self.pandaActor.posInterval(13,
                                                   Point3(0, 10, 0),
                                                   startPos=Point3(0, -10, 0))
        hprInterval1 = self.pandaActor.hprInterval(3, Point3(180, 0, 0), startHpr=Point3(0, 0.0))
        hprInterval2 = self.pandaActor.hprInterval(3, Point3(0, 0, 0), startHpr=Point3(180, 0, 0))

        self.pandaPace = Sequence(posInterval1, hprInterval1,
                                  posInterval2, hprInterval2,
                                  name="pandaPace")
        self.pandaPace.loop()
        self.dlight = DirectionalLight('my dlight')
        dlnp = self.render.attachNewNode(self.dlight)
        self.pandaActor.setLight(dlnp)
        self.render.setLight(dlnp)

        ambientLight = AmbientLight('ambientLight')
        ambientLight.setColor((0.1, 0.1, 0.1, 1))
        ambientLightNP = self.render.attachNewNode(ambientLight)
        self.render.setLight(ambientLightNP)

        directionalLight = DirectionalLight('directionalLight')
        directionalLight.setColor((0.8, 1, 1, 1))
        directionalLight.setShadowCaster(True, 512, 512)
        directionalLightNp = self.render.attachNewNode(directionalLight)
        directionalLightNp.setHpr(180, -20, 0)
        self.render.setLight(directionalLightNp)
        self.render.setShaderAuto()

        ambient = AmbientLight('ambient')
        ambient.setColor((0.5, 1, 0.5, 1))
        ambientNP = self.pandaActor.attachNewNode(ambient)

        self.pandaActor.setLightOff()
        self.pandaActor.setLight(ambientNP)

        self.useDrive()  # 鼠标键盘控制模式，很难用
        self.useTrackball()  # 默认的控制模式
        bk_text = "This is my Demo"
        # self.textObject = OnscreenText(text=bk_text, pos=(0.95, -0.95), scale=0.07,
        #                                fg=(1, 0.5, 0.5, 1), align=TextNode.ACenter,
        #                                mayChange=1)
        # b = DirectButton(text="OK",
        #                  scale=0.2, command=self.setText)

    def setText(self):
        bk_text = "Button Clicked"
        self.textObject.setText(bk_text)

    def spinCameraTask(self, task):
        angleDegrees = task.time * 6.0
        angleRadians = angleDegrees * (pi / 190.0)
        self.camera.setPos(20 * sin(angleRadians), -20 * cos(angleRadians), 3)
        self.camera.setHpr(angleDegrees, 0, 0)
        return Task.cont


app = MyApp()
app.run()
